/*
 *  a2, an Apple II emulator in C
 *  (c) Copyright 1990 by Rich Skrenta
 *
 *  Command line interface written by Tom Markson
 *
 *  Distribution agreement:
 *
 *	You may freely copy or redistribute this software, so long
 *	as there is no profit made from its use, sale, trade or
 *	reproduction.  You may not change this copyright notice,
 *	and it must be included prominently in any copy made.
 *
 *  Send emulator related mail to:  skrenta@blekko.commodore.com
 *				    skrenta@blekko.uucp
 *
 *  Modified by Chris Osborn <osborn@athena.ecs.csus.edu>
 */

#define		TRUE		1
#define		FALSE		0

extern FILE	*sfp;

#define	assert(p)	if(! (p)) asfail(__FILE__, __LINE__, "p"); else


#define		join(low, high)		((low) | ((high) << 8))
#define		low(x)			((x) & 0xFF)
#define		high(x)			((x) >> 8)

#define		tobinary(v)		((v >> 4) * 10 + (v & 0xF))
#define		tobcd(v)		((v % 10) | ((v / 10) << 4))

#define		push(s)			WRITELOC((0x100 | Sp--), s);
#define		pop()			READLOC(0x100 | ++Sp)

#define		SIGN(x)			N = ((x) & 0x80)
#define		ZERO(x)			Z = (x)
#define		CARRY(x)		C = ((x) > 0xFF)
#define		dCARRY(x)		C = ((x) > 99)


#define	REF(a)		READLOC(a)
#define	REFptr		READLOC(ptr)
#define	REFzero		READLOC(ptr)
#define	REFimm		READLOC(Pc++)

#define	SET(a,v)	WRITELOC(a,v)
#define	SETptr(v)	WRITELOC(ptr,v)
#define	SETzero(v)	WRITELOC(ptr, v);

#define	PCINC		Pc++;

#define	ABSOL		ptr = READLOC(Pc++); ptr |= READLOC(Pc++) << 8;
#define	ABSX		ptr = READLOC(Pc++); ptr |= READLOC(Pc++) << 8; ptr += X;
#define	ABSY		ptr = READLOC(Pc++); ptr |= READLOC(Pc++) << 8; ptr += Y;
#define	ZEROP		ptr = READLOC(Pc++);
#define	ZEROX		ptr = (READLOC(Pc++) + X) & 0xFF;
#define	ZEROY		ptr = (READLOC(Pc++) + Y) & 0xFF;
#define INDZP		stmp = READLOC(Pc++); ptr = READLOC(stmp++); ptr |= READLOC(stmp) << 8;
#define	INDX		stmp = READLOC(Pc++) + X; ptr = READLOC(stmp++); ptr |= READLOC(stmp) << 8;
#define	INDY		stmp = READLOC(Pc++); ptr = READLOC(stmp++); ptr |= READLOC(stmp) << 8; ptr += Y;

#define	rZEROP		READLOC( READLOC(Pc++) )
#define	rZEROX		READLOC( (READLOC(Pc++) + X) & 0xFF )
#define	rZEROY		READLOC( (READLOC(Pc++) + Y) & 0xFF )
#define	sZEROP(v)	WRITELOC( READLOC(Pc++), v)
#define	sZEROX(v)	WRITELOC( (READLOC(Pc++) + X) & 0xFF, v);
#define	sZEROY(v)	WRITELOC( (READLOC(Pc++) + Y) & 0xFF, v);

#define	ZIND		ptr = READLOC( READLOC(Pc++) );
#define	ABINDX		ptmp = READLOC(Pc++); ptmp |= READLOC(Pc++) << 8; ptmp += X; ptr = READLOC(ptmp++); ptr |= READLOC(ptmp) << 8;


/*
 *  Since we store flags as 0/nonzero instead of 0/1, we need to turn
 *  nonzero back into 1.  Thus, "!!" is not a no-op.
 */

#define	get_status()   ((!!N << 7)|(!!V << 6)|(!!B << 4)|(!!D << 3)|(!!I << 2)|(!Z << 1)|(!!C))

extern unsigned short Pc;
extern unsigned char Sp;	/* stack pointer 	*/
extern unsigned int A;		/* accumulator		*/
extern unsigned char X;		/* X register		*/
extern unsigned char Y;		/* Y register		*/

extern unsigned int N;		/* 7 - sign		*/
extern unsigned int V;		/* 6 - Overflow		*/
				/* 5 - Unused		*/
extern unsigned int B;		/* 4 - Break		*/
extern unsigned int D;		/* 3 - Decimal		*/
extern unsigned int I;		/* 2 - Interrupt	*/
extern unsigned int Z;		/* 1 - inverse of Zero  */
extern unsigned int C;		/* 0 - Carry		*/


/*
 *  Note:
 *	Flags are stored as zero/nonzero instead of 0/1, except for
 *	N which is stored 0/0x80.  Z is "not Z"; use !Z to get
 *	the zero flag.
 *
 *	The flags code is tight, especially around overflow set on ADC
 *	and SBC.  Integer basic is a good test to see if it still works.
 */


#define	DO_RTS	Pc = pop(); Pc |= pop() << 8; Pc++;

