/*
 * Controller.m - does all initialization and handles
 *		  control panel
 * written by Chris Osborn <osborn@athena.ecs.csus.edu>
 */

#import "Controller.h"
#import "apple.h"
#import "VideoView.h"
#import "CPU.h"
#import <appkit/ScrollView.h>
#import <appkit/Matrix.h>
#import <appkit/ButtonCell.h>
#import <appkit/NXImage.h>
#import <appkit/Panel.h>
#import <appkit/OpenPanel.h>
#import <appkit/Text.h>
#import <appkit/PopUpList.h>
#import <appkit/Button.h>
#import <objc/List.h>
#import <defaults.h>
#import <strings.h>
#import <sys/types.h>
#import <sys/stat.h>
#import <sys/dir.h>
#import <sys/param.h>

#define APPNAME "Diamond"

#define DISK1NAME "PRODOS.IMAGE.D1"
#define DISK2NAME "PRODOS.IMAGE.D2"

#define KEY_LA 0
#define KEY_RA 1
#define KEY_LC 2
#define KEY_RC 3

static char *YesNo[] = {"No", "Yes"};

NXDefaultsVector IIe_Defaults = {
  {"ProdosDisk1", ""},
  {"ProdosDisk2", ""},
  {"OptionKey", "LA"},
  {"AppleKey", "RA"},
  {"BufferKeys", "No"},
  {"UnixVolume", ""},
  {NULL}
};

char appHome[MAXPATHLEN];
extern int bufferKeys;
extern char unixVolumePath[MAXPATHLEN];
extern char unixVolumeName[MAXPATHLEN];

@implementation Controller

- setDiskList1:anObject
{
  id matrix = [[anObject target] itemList];


  [matrix setTarget:self];
  [matrix setAction:@selector(changeDisk:)];
  [matrix setTag:2];
  diskList1 = anObject;

  return self;
}

- setDiskList2:anObject
{
  id matrix = [[anObject target] itemList];


  [matrix setTarget:self];
  [matrix setAction:@selector(changeDisk:)];
  [matrix setTag:3];
  diskList2 = anObject;

  return self;
}

- appDidInit:sender
{
  NXRegisterDefaults (APPNAME, IIe_Defaults);
  
  [self loadFiles];
  [self initPanel];
  [self setUpKeyboard];
  [videoScreen initdisplay];
  [cpu powerUp];

  [cpu execute];
  
  return self;
}

- setUpKeyboard
{
  char *Ap, *Op;
  int ab, ob;
  int t;
    

  Ap = (char *) NXGetDefaultValue (APPNAME, "BufferKeys");
  bufferKeys = Ap[0] == 'Y';
  [keyBufButton setState:bufferKeys];
  
  Ap = (char *) NXGetDefaultValue (APPNAME, "AppleKey");
  Op = (char *) NXGetDefaultValue (APPNAME, "OptionKey");
  
  ab = (Ap[0] == 'R') + (Ap[1] == 'C') * 2;
  ob = (Op[0] == 'R') + (Op[1] == 'C') * 2;
  
  switch (ab) {
  case KEY_LA:
    AppleKeyBit = NX_NEXTLALTKEYMASK;
    break;
  case KEY_RA:
    AppleKeyBit = NX_NEXTRALTKEYMASK;
    break;
  case KEY_LC:
    AppleKeyBit = NX_NEXTLCMDKEYMASK;
    break;
  case KEY_RC:
    AppleKeyBit = NX_NEXTRCMDKEYMASK;
    break;
  }
  
  switch (ob) {
  case KEY_LA:
    OptionKeyBit = NX_NEXTLALTKEYMASK;
    break;
  case KEY_RA:
    OptionKeyBit = NX_NEXTRALTKEYMASK;
    break;
  case KEY_LC:
    OptionKeyBit = NX_NEXTLCMDKEYMASK;
    break;
  case KEY_RC:
    OptionKeyBit = NX_NEXTRCMDKEYMASK;
    break;
  }

  t = 0;
  if (ab == KEY_LC && ob == KEY_LA)
    t = 1;
  else if (ab == KEY_LA && ob == KEY_RA)
    t = 2;
  else if (ab == KEY_RC && ob == KEY_RA)
    t = 3;
  
  [keyMap selectCellAt:t :0];
  
  return self;
}

- loadFiles
{
  FILE *romfile;
  char *s;
  char buf[MAXPATHLEN];
  id alist1, alist2;
  int i, j;
  char *home_dir;
  char *p;
  DIR *dir;
  struct direct *entry;
  char *getenv ();
  

  home_dir = getenv ("HOME");

  strcpy (appHome, NXArgv[0]);
  if (s = rindex (appHome, '/')) {
    s++;
    *s = 0;
  }
  else
    appHome[0] = 0;
  
  sprintf (buf, "%sEK2EROM", appHome);
  romfile = fopen (buf, "r");
  fread (&SysMem[0xc100 + ROM], 0x3f00, 1, romfile);
  
  fclose (romfile);

  alist1 = [diskList1 target];
  alist2 = [diskList2 target];
  i = [alist1 count];
  for (j = 0; j < i; j++)
    [alist1 removeItemAt:j];
  i = [alist2 count];
  for (j = 0; j < i; j++)
    [alist2 removeItemAt:j];

  if (!strlen (NXGetDefaultValue (APPNAME, "ProdosDisk1"))) {
    sprintf (buf, "%s%s", appHome, DISK1NAME);
    [alist1 addItem:buf];
  }
  else
    [alist1 addItem:NXGetDefaultValue (APPNAME, "ProdosDisk1")];
  [diskList1 setTitle:[[[alist1 itemList] cellAt:0:0] title]];
  
  if (!strlen (NXGetDefaultValue (APPNAME, "ProdosDisk2"))) {
    sprintf (buf, "%s%s", appHome, DISK2NAME);
    [alist2 addItem:buf];
  }
  else
    [alist2 addItem:NXGetDefaultValue (APPNAME, "ProdosDisk2")];
  [diskList2 setTitle:[[[alist2 itemList] cellAt:0:0] title]];

  if (!strlen (NXGetDefaultValue (APPNAME, "UnixVolume"))) {
    [unixField setStringValue:home_dir];
  }
  else
    [unixField setStringValue:NXGetDefaultValue (APPNAME, "UnixVolume")];

  strcpy (unixVolumePath, [unixField stringValue]);
  if (p = rindex (unixVolumePath, '/')) {
    strcpy (unixVolumeName, p+1);
    *p = 0;
  }
  else {
    strcpy (unixVolumeName, unixVolumePath);
    unixVolumePath[0] = 0;
  }  
  
  sprintf (buf, "%s%s", appHome, DISK1NAME);
  [alist1 addItem:buf];
  [alist2 addItem:buf];
  sprintf (buf, "%s%s", appHome, DISK2NAME);
  [alist1 addItem:buf];
  [alist2 addItem:buf];
  
  if (dir = opendir ("/LocalLibrary/AppleDisks")) {
    while (entry = readdir (dir))
      if (strcmp (entry->d_name, ".") && strcmp (entry->d_name, "..")) {
	sprintf (buf, "/LocalLibrary/AppleDisks/%s", entry->d_name);
	[alist1 addItem:buf];
	[alist2 addItem:buf];
      }
    closedir(dir);
  }

  sprintf (buf, "%s/Library/AppleDisks", home_dir);
  if (dir = opendir (buf)) {
    while (entry = readdir (dir))
      if (strcmp (entry->d_name, ".") && strcmp (entry->d_name, "..")) {
	sprintf (buf, "%s/Library/AppleDisks/%s", home_dir, entry->d_name);
	[alist1 addItem:buf];
	[alist2 addItem:buf];
      }
    closedir(dir);
  }

  ProInit ([diskList1 title], [diskList2 title]);

  return self;
}

- initPanel
{
  NXSize space;


  [[myProto cell] setShowsStateBy:NX_CHANGEBACKGROUND];
  [[myProto cell] setHighlightsBy:NX_CHANGEBACKGROUND];
  
  optionButtons = [[Matrix allocFromZone:[self zone]] initFrame:NULL mode:NX_RADIOMODE prototype:[myProto cell] numRows:1 numCols:3];
  space.width = 64;
  space.height = 64;
  [optionButtons setCellSize:&space];
  space.width = 0;
  space.height = 0;
  [optionButtons setIntercell:&space];
  [optionButtons sizeToCells];
  [optionButtons setAutoscroll:YES];
  [optionButtons setAction:@selector(chooseOption:)];
  [optionButtons setTarget:self];

  [[optionButtons cellAt:0 :0] setImage:[NXImage findImageNamed:"Floppy"]];
  [[optionButtons cellAt:0 :1] setImage:[NXImage findImageNamed:"Keyboard"]];
  [[optionButtons cellAt:0 :2] setImage:[NXImage findImageNamed:"Unix"]];
  
  [optionScroll setBackgroundGray:NX_LTGRAY];
  [optionScroll setBorderType:NX_BEZEL];
  [optionScroll setHorizScrollerRequired:YES];
  [[optionScroll setDocView:optionButtons] free];

  optionList = [[List allocFromZone:[self zone]] init];
  [optionList insertObject:floppyView at:0];
  [optionList insertObject:keyboardView at:1];
  [optionList insertObject:unixView at:2];

  [[optionView contentView] addSubview:[optionList objectAt:0]];
  lastOpView = [optionList objectAt:0];
  
  return self;
}

- chooseOption:sender
{
  id newView;
  
  
  if ((newView = [optionList objectAt:[sender selectedCol]]) && newView != lastOpView) {
    [[optionView contentView] replaceSubview: lastOpView with:newView];
    lastOpView = newView;
  }
  
  [optionView display];

  return self;
}

- switchDisk:(char *) diskName inDrive:(int) drive
{
  struct stat buf;


  if (stat (diskName, &buf)) {
    NXRunAlertPanel ("Alert", "Cannot open %s", "Ok", NULL, NULL, diskName);
    return nil;
  }
  else {
    ProChangeDisk (drive, diskName);
    if (drive)
      NXWriteDefault (APPNAME, "ProdosDisk2", diskName);
    else
      NXWriteDefault (APPNAME, "ProdosDisk1", diskName);
  }
  
  return self;
}

- changeDisk:sender
{
  id buttonObject;
  id myOP;
  const char *const types[] = {NULL};
  

  if ([sender tag] > 1) {
    [self switchDisk:(char *) [[sender selectedCell] title] inDrive:[sender tag] % 2];
  }
  else {
    buttonObject = [sender tag] ? diskList2 : diskList1;

    myOP = [OpenPanel new];

    [myOP setTitle:"Choose disk"];
    if ([myOP runModalForTypes:types]) {
      [buttonObject setTitle:[myOP filename]];
      [self switchDisk:(char *) [buttonObject title] inDrive:[sender tag]];
    }
  }

  return self;
}

- changeKeyBuffer:sender
{
  bufferKeys = [sender state];
  NXWriteDefault (APPNAME, "BufferKeys", YesNo[bufferKeys]);
  
  return self;
}

- changeKeyMap:sender
{
  static char *whichKey[] = {"LA", "RA", "LC", "RC"};
  int t;
  int ab = 0, ob = 0;
  

  t = [[sender selectedCell] tag];
  
  switch (t) {
  case 0:
    ab = KEY_RA;
    ob = KEY_LA;
    break;
  case 1:
    ab = KEY_LC;
    ob = KEY_LA;
    break;
  case 2:
    ab = KEY_LA;
    ob = KEY_RA;
    break;
  case 3:
    ab = KEY_RC;
    ob = KEY_RA;
    break;
  }

  NXWriteDefault (APPNAME, "AppleKey", whichKey[ab]);
  NXWriteDefault (APPNAME, "OptionKey", whichKey[ob]);

  switch (ab) {
  case KEY_LA:
    AppleKeyBit = NX_NEXTLALTKEYMASK;
    break;
  case KEY_RA:
    AppleKeyBit = NX_NEXTRALTKEYMASK;
    break;
  case KEY_LC:
    AppleKeyBit = NX_NEXTLCMDKEYMASK;
    break;
  case KEY_RC:
    AppleKeyBit = NX_NEXTRCMDKEYMASK;
    break;
  }
  
  switch (ob) {
  case KEY_LA:
    OptionKeyBit = NX_NEXTLALTKEYMASK;
    break;
  case KEY_RA:
    OptionKeyBit = NX_NEXTRALTKEYMASK;
    break;
  case KEY_LC:
    OptionKeyBit = NX_NEXTLCMDKEYMASK;
    break;
  case KEY_RC:
    OptionKeyBit = NX_NEXTRCMDKEYMASK;
    break;
  }

  return self;
}

- loadFile:sender
{
  id myOP;
  const char *const types[] = {NULL};
  FILE *file;
  int i;
  

  myOP = [OpenPanel new];

  [myOP setTitle:"Choose file"];
  if ([myOP runModalForTypes:types]) {
    file = fopen ([myOP filename], "r");
    i = 0;
    while (fread (&SysMem[0x2000 + i], 1, 1, file))
      i++;
    NXRunAlertPanel ("File loaded", "%s loaded at 0x2000, length 0x%04X", "Ok", NULL, NULL, [myOP filename], i);
  }

  return self;
}

- setUnixVol:sender
{
  id myOP;
  const char *const types[] = {NULL};
  char *p;
  

  myOP = [OpenPanel new];
  [myOP setDelegate:self];

  [myOP setTitle:"Choose directory"];
  if ([myOP runModalForTypes:types]) {
    [unixField setStringValue:[myOP filename]];
    strcpy (unixVolumePath, [unixField stringValue]);
    if (p = rindex (unixVolumePath, '/')) {
      strcpy (unixVolumeName, p+1);
      *p = 0;
    }
    else {
      strcpy (unixVolumeName, unixVolumePath);
      unixVolumePath[0] = 0;
    }
  }
  [myOP setDelegate:nil];
  
  return self;
}

-(BOOL) panel:sender filterFile:(const char *) filename inDirectory:(const char *) directory
{
  struct stat sbuf;
  static char buf[MAXPATHLEN];
  

  sprintf (buf, "%s/%s", directory, filename);
  stat (buf, &sbuf);
  if (sbuf.st_mode & S_IFDIR)
    return YES;

  return NO;
}

-(BOOL) panelValidateFilenames:sender
{
  fprintf (stderr, "Supposed to validate something\n");

  return YES;
}

@end
