/*
 * Debugger.m - has monitor like commands available
 *		from within the debugger
 * written by Chris Osborn <osborn@athena.ecs.csus.edu>
 * some code copied from Ben Koning <ben@apple.com>
 */


#import "Debugger.h"
#import "apple.h"
#import <appkit/Window.h>
#import <appkit/Font.h>
#import <appkit/Text.h>
#import <cthreads.h>
#import <strings.h>

extern int Debugging;
extern condition_t debugStep;
extern condition_t debugStepped;
extern mutex_t debugLock;
int BreakPoint;

@implementation Debugger

-(char *) disasm:(int) Pc bytes:(int *) len
{
  int opcode = READLOC (Pc);
  int n;
  char *mne;
  static char buf[40];
  

  switch (opcode) {
  case 0x69:			/* ADC #dd */
    mne = "ADC #%02X"; n = 1; 
    break;
  case 0x65:			/* ADC aa */
    mne = "ADC %02X"; n = 1; 
    break;
  case 0x75:			/* ADC aa,X */
    mne = "ADC %02X,X"; n = 1; 
    break;
  case 0x6d:			/* ADC aaaa */
    mne = "ADC %04X"; n = 2; 
    break;
  case 0x7d:			/* ADC aaaa,X */
    mne = "ADC %04X,X"; n = 2; 
    break;
  case 0x79:			/* ADC aaaa,Y */
    mne = "ADC %04X,Y"; n = 2; 
    break;
  case 0x61:			/* ADC (aa,X) */
    mne = "ADC (%02X,X)"; n = 1; 
    break;
  case 0x71:			/* ADC (aa),Y */
    mne = "ADC (%02X),Y"; n = 1; 
    break;
  case 0x72:			/* ADC (aa) */
    mne = "ADC (%02X)"; n = 1;
    break;
  case 0x29:			/* AND #dd */
    mne = "AND #%02X"; n = 1; 
    break;
  case 0x25:			/* AND aa */
    mne = "AND %02X"; n = 1; 
    break;
  case 0x35:			/* AND aa,X */
    mne = "AND %02X,X"; n = 1; 
    break;
  case 0x2d:			/* AND aaaa */
    mne = "AND %04X"; n = 2; 
    break;
  case 0x3d:			/* AND aaaa,X */
    mne = "AND %04X,X"; n = 2; 
    break;
  case 0x39:			/* AND aaaa,Y */
    mne = "AND %04X,Y"; n = 2; 
    break;
  case 0x32:			/* AND (aa) */
    mne = "AND (%02X)"; n = 1;
    break;
  case 0x21:			/* AND (aa,X) */
    mne = "AND (%02X,X)"; n = 1; 
    break;
  case 0x31:			/* AND (aa),Y */
    mne = "AND (%02X),Y"; n = 1; 
    break;
  case 0x0a:			/* ASL A */
    mne = "ASL"; n = 0; 
    break;
  case 0x06:			/* ASL aa */
    mne = "ASL %02X"; n = 1; 
    break;
  case 0x16:			/* ASL aa,X */
    mne = "ASL %02X,X"; n = 1; 
    break;
  case 0x0e:			/* ASL aaaa */
    mne = "ASL %04X"; n = 2; 
    break;
  case 0x1e:			/* ASL aaaa,X */
    mne = "ASL %04X,X"; n = 2; 
    break;
  case 0x90:			/* BCC rr */
    mne = "BCC %02X"; n = -1; 
    break;
  case 0xb0:			/* BCS rr */
    mne = "BCS %02X"; n = -1; 
    break;
  case 0xf0:			/* BEQ rr */
    mne = "BEQ %02X"; n = -1; 
    break;
  case 0x24:			/* BIT aa */
    mne = "BIT %02X"; n = 1; 
    break;
  case 0x2c:			/* BIT aaaa */
    mne = "BIT %04X"; n = 2; 
    break;
  case 0x34:			/* BIT aa,X */
    mne = "BIT %02X"; n = 1;
    break;
  case 0x3c:			/* BIT aaaa,X */
    mne = "BIT %04X,X"; n = 2;
    break;
  case 0x30:			/* BMI rr */
    mne = "BMI %02X"; n = -1; 
    break;
  case 0xd0:			/* BNE rr */
    mne = "BNE %02X"; n = -1; 
    break;
  case 0x10:			/* BPL rr */
    mne = "BPL %02X"; n = -1; 
    break;
  case 0x80:			/* BRA */
    mne = "BRA %02X"; n = -1;
    break;
  case 0x00:			/* BRK */
    mne = "BRK %02X"; n = 1; 
    break;
  case 0x50:			/* BVC rr */
    mne = "BVC %02X"; n = -1; 
    break;
  case 0x70:			/* BVS rr */
    mne = "BVS %02X"; n = -1; 
    break;
  case 0x18:			/* CLC */
    mne = "CLC"; n = 0; 
    break;
  case 0xd8:			/* CLD */
    mne = "CLD"; n = 0; 
    break;
  case 0x58:			/* CLI */
    mne = "CLI"; n = 0; 
    break;
  case 0xb8:			/* CLV */
    mne = "CLV"; n = 0; 
    break;
  case 0xc9:			/* CMP #dd */
    mne = "CMP #%02X"; n = 1; 
    break;
  case 0xc5:			/* CMP aa */
    mne = "CMP %02X"; n = 1; 
    break;
  case 0xd5:			/* CMP aa,X */
    mne = "CMP %02X,X"; n = 1; 
    break;
  case 0xcd:			/* CMP aaaa */
    mne = "CMP %04X"; n = 2; 
    break;
  case 0xdd:			/* CMP aaaa,X */
    mne = "CMP %04X,X"; n = 2; 
    break;
  case 0xd9:			/* CMP aaaa,Y */
    mne = "CMP %04X,Y"; n = 2; 
    break;
  case 0xc1:			/* CMP (aa,X) */
    mne = "CMP (%02X,X)"; n = 1; 
    break;
  case 0xd1:			/* CMP (aa),Y */
    mne = "CMP (%02X),Y"; n = 1; 
    break;
  case 0xd2:			/* CMP (aa) */
    mne = "CMP (%02X)"; n = 1;
    break;
  case 0xe0:			/* CPX #dd */
    mne = "CPX #%02X"; n = 1; 
    break;
  case 0xe4:			/* CPX aa */
    mne = "CPX %02X"; n = 1; 
    break;
  case 0xec:			/* CPX aaaa */
    mne = "CPX %04X"; n = 2; 
    break;
  case 0xc0:			/* CPY #dd */
    mne = "CPY #%02X"; n = 1; 
    break;
  case 0xc4:			/* CPY aa */
    mne = "CPY %02X"; n = 1; 
    break;
  case 0xcc:			/* CPY aaaa */
    mne = "CPY %04X"; n = 2; 
    break;
  case 0x3a:			/* DEA */
    mne = "DEA"; n = 0;
    break;
  case 0xc6:			/* DEC aa */
    mne = "DEC %02X"; n = 1; 
    break;
  case 0xd6:			/* DEC aa,X */
    mne = "DEC %02X,X"; n = 1; 
    break;
  case 0xce:			/* DEC aaaa */
    mne = "DEC %04X"; n = 2; 
    break;
  case 0xde:			/* DEC aaaa,X */
    mne = "DEC %04X,X"; n = 2; 
    break;
  case 0xca:			/* DEX */
    mne = "DEX"; n = 0; 
    break;
  case 0x88:			/* DEY */
    mne = "DEY"; n = 0; 
    break;
  case 0x49:			/* EOR #dd */
    mne = "EOR #%02X"; n = 1; 
    break;
  case 0x45:			/* EOR aa */
    mne = "EOR %02X"; n = 1; 
    break;
  case 0x55:			/* EOR aa,X */
    mne = "EOR %02X,X"; n = 1; 
    break;
  case 0x4d:			/* EOR aaaa */
    mne = "EOR %04X"; n = 2; 
    break;
  case 0x5d:			/* EOR aaaa,X */
    mne = "EOR %04X,X"; n = 2; 
    break;
  case 0x59:			/* EOR aaaa,Y */
    mne = "EOR %04X,Y"; n = 2; 
    break;
  case 0x41:			/* EOR (aa,X) */
    mne = "EOR (%02X,X)"; n = 1; 
    break;
  case 0x51:			/* EOR (aa),Y */
    mne = "EOR (%02X),Y"; n = 1; 
    break;
  case 0x52:			/* EOR (aa) */
    mne = "EOR %02X", n = 1;
    break;
  case 0x1a:			/* INA */
    mne = "INC"; n = 0;
    break;
  case 0xe6:			/* INC aa */
    mne = "INC %02X"; n = 1; 
    break;
  case 0xf6:			/* INC aa,X */
    mne = "INC %02X,X"; n = 1; 
    break;
  case 0xee:			/* INC aaaa */
    mne = "INC %04X"; n = 2; 
    break;
  case 0xfe:			/* INC aaaa,X */
    mne = "INC %04X,X"; n = 2; 
    break;
  case 0xe8:			/* INX */
    mne = "INX"; n = 0; 
    break;
  case 0xc8:			/* INY */
    mne = "INY"; n = 0; 
    break;
  case 0x4c:			/* JMP aaaa */
    mne = "JMP %04X"; n = 2; 
    break;
  case 0x6c:			/* JMP (aaaa) */
    mne = "JMP (%04X)"; n = 2; 
    break;
  case 0x7c:			/* JMP (aaaa,X) */
    mne = "JMP (%04X,X)"; n = 2;
    break;
  case 0x20:			/* JSR aaaa */
    mne = "JSR %04X"; n = 2; 
    break;
  case 0xa9:			/* LDA #dd */
    mne = "LDA #%02X"; n = 1; 
    break;
  case 0xa5:			/* LDA aa */
    mne = "LDA %02X"; n = 1; 
    break;
  case 0xb5:			/* LDA aa,X */
    mne = "LDA %02X,X"; n = 1; 
    break;
  case 0xad:			/* LDA aaaa */
    mne = "LDA %04X"; n = 2; 
    break;
  case 0xbd:			/* LDA aaaa,X */
    mne = "LDA %04X,X"; n = 2; 
    break;
  case 0xb9:			/* LDA aaaa,Y */
    mne = "LDA %04X,Y"; n = 2; 
    break;
  case 0xa1:			/* LDA (aa,X) */
    mne = "LDA (%02X,X)"; n = 1; 
    break;
  case 0xb1:			/* LDA (aa),Y */
    mne = "LDA (%02X),Y"; n = 1; 
    break;
  case 0xb2:			/* LDA (aa) */
    mne = "LDA (%02X)"; n = 1;
    break;
  case 0xa2:			/* LDX #dd */
    mne = "LDX #%02X"; n = 1; 
    break;
  case 0xa6:			/* LDX aa */
    mne = "LDX %02X"; n = 1; 
    break;
  case 0xb6:			/* LDX aa,Y */
    mne = "LDX %02X,Y"; n = 1; 
    break;
  case 0xae:			/* LDX aaaa */
    mne = "LDX %04X"; n = 2; 
    break;
  case 0xbe:			/* LDX aaaa,Y */
    mne = "LDX %04X,Y"; n = 2; 
    break;
  case 0xa0:			/* LDY #dd */
    mne = "LDY #%02X"; n = 1; 
    break;
  case 0xa4:			/* LDY aa */
    mne = "LDY %02X"; n = 1; 
    break;
  case 0xb4:			/* LDY aa,X */
    mne = "LDY %02X,X"; n = 1; 
    break;
  case 0xac:			/* LDY aaaa */
    mne = "LDY %04X"; n = 2; 
    break;
  case 0xbc:			/* LDY aaaa,X */
    mne = "LDY %04X,X"; n = 2; 
    break;
  case 0x4a:			/* LSR A */
    mne = "LSR"; n = 0; 
    break;
  case 0x46:			/* LSR aa */
    mne = "LSR %02X"; n = 1; 
    break;
  case 0x56:			/* LSR aa,X */
    mne = "LSR %02X,X"; n = 1; 
    break;
  case 0x4e:			/* LSR aaaa */
    mne = "LSR %04X"; n = 2; 
    break;
  case 0x5e:			/* LSR aaaa,X */
    mne = "LSR %04X,X"; n = 2; 
    break;
  case 0xea:			/* NOP */
    mne = "NOP"; n = 0; 
    break;
  case 0x09:			/* ORA #dd */
    mne = "ORA #%02X"; n = 1; 
    break;
  case 0x05:			/* ORA aa */
    mne = "ORA %02X"; n = 1; 
    break;
  case 0x15:			/* ORA aa,X */
    mne = "ORA %02X,X"; n = 1; 
    break;
  case 0x0d:			/* ORA aaaa */
    mne = "ORA %04X"; n = 2; 
    break;
  case 0x1d:			/* ORA aaaa,X */
    mne = "ORA %04X,X"; n = 2; 
    break;
  case 0x19:			/* ORA aaaa,Y */
    mne = "ORA %04X,Y"; n = 2; 
    break;
  case 0x01:			/* ORA (aa,X) */
    mne = "ORA (%02X,X)"; n = 1; 
    break;
  case 0x11:			/* ORA (aa),Y */
    mne = "ORA (%02X),Y"; n = 1; 
    break;
  case 0x12:			/* ORA (aa) */
    mne = "ORA (%02X)"; n = 1;
    break;
  case 0x48:			/* PHA */
    mne = "PHA"; n = 0; 
    break;
  case 0x08:			/* PHP */
    mne = "PHP"; n = 0; 
    break;
  case 0xda:			/* PHX */
    mne = "PHX"; n = 0;
    break;
  case 0x5a:			/* PHY */
    mne = "PHY"; n = 0;
    break;
  case 0x68:			/* PLA */
    mne = "PLA"; n = 0; 
    break;
  case 0x28:			/* PLP */
    mne = "PLP"; n = 0; 
    break;
  case 0xfa:			/* PLX */
    mne = "PLX"; n = 0;
    break;
  case 0x7a:			/* PLY */
    mne = "PLY"; n = 0;
    break;
  case 0x2a:			/* ROL A */
    mne = "ROL"; n = 0; 
    break;
  case 0x26:			/* ROL aa */
    mne = "ROL %02X"; n = 1; 
    break;
  case 0x36:			/* ROL aa,X */
    mne = "ROL %02X,X"; n = 1; 
    break;
  case 0x2e:			/* ROL aaaa */
    mne = "ROL %04X"; n = 2; 
    break;
  case 0x3e:			/* ROL aaaa,X */
    mne = "ROL %04X,X"; n = 2; 
    break;
  case 0x6a:			/* ROR A */
    mne = "ROR"; n = 0; 
    break;
  case 0x66:			/* ROR aa */
    mne = "ROR %02X"; n = 1; 
    break;
  case 0x76:			/* ROR aa,X */
    mne = "ROR %02X,X"; n = 1; 
    break;
  case 0x6e:			/* ROR aaaa */
    mne = "ROR %04X"; n = 2; 
    break;
  case 0x7e:			/* ROR aaaa,X */
    mne = "ROR %04X,X"; n = 2; 
    break;
  case 0x40:			/* RTI */
    mne = "RTI"; n = 0; 
    break;
  case 0x60:			/* RTS */
    mne = "RTS"; n = 0; 
    break;
  case 0xe9:			/* SBC #dd */
    mne = "SBC #%02X"; n = 1; 
    break;
  case 0xe5:			/* SBC aa */
    mne = "SBC %02X"; n = 1; 
    break;
  case 0xf5:			/* SBC aa,x */
    mne = "SBC %02X,X"; n = 1; 
    break;
  case 0xed:			/* SBC aaaa */
    mne = "SBC %04X"; n = 2; 
    break;
  case 0xfd:			/* SBC aaaa,X */
    mne = "SBC %04X,X"; n = 2; 
    break;
  case 0xf9:			/* SBC aaaa,Y */
    mne = "SBC %04X,Y"; n = 2; 
    break;
  case 0xe1:			/* SBC (aa,X) */
    mne = "SBC (%02X,X)"; n = 1; 
    break;
  case 0xf1:			/* SBC (aa),Y */
    mne = "SBC (%02X),Y"; n = 1; 
    break;
  case 0xf2:			/* SBC (aa) */
    mne = "SBC (%02X)"; n = 1;
    break;
  case 0x38:			/* SEC */
    mne = "SEC"; n = 0; 
    break;
  case 0xf8:			/* SED */
    mne = "SED"; n = 0; 
    break;
  case 0x78:			/* SEI */
    mne = "SEI"; n = 0; 
    break;
  case 0x85:			/* STA aa */
    mne = "STA %02X"; n = 1; 
    break;
  case 0x95:			/* STA aa,X */
    mne = "STA %02X,X"; n = 1; 
    break;
  case 0x8d:			/* STA aaaa */
    mne = "STA %04X"; n = 2; 
    break;
  case 0x9d:			/* STA aaaa,X */
    mne = "STA %04X,X"; n = 2; 
    break;
  case 0x99:			/* STA aaaa,Y */
    mne = "STA %04X,Y"; n = 2; 
    break;
  case 0x81:			/* STA (aa,X) */
    mne = "STA (%02X,X)"; n = 1; 
    break;
  case 0x92:			/* STA (aa) */
    mne = "STA (%02X)"; n = 1;
    break;
  case 0x91:			/* STA (aa),Y */
    mne = "STA (%02X),Y"; n = 1; 
    break;
  case 0x86:			/* STX aa */
    mne = "STX %02X"; n = 1; 
    break;
  case 0x96:			/* STX aa,Y */
    mne = "STX %02X,Y"; n = 1; 
    break;
  case 0x8e:			/* STX aaaa */
    mne = "STX %04X"; n = 2; 
    break;
  case 0x84:			/* STY aa */
    mne = "STY %02X"; n = 1; 
    break;
  case 0x94:			/* STY aa,X */
    mne = "STY %02X,X"; n = 1; 
    break;
  case 0x8c:			/* STY aaaa */
    mne = "STY %04X"; n = 2; 
    break;
  case 0x64:			/* STZ aa */
    mne = "STZ %02X"; n = 1;
    break;
  case 0x74:			/* STZ aa,X */
    mne = "STZ %02X,X"; n = 1;
    break;
  case 0x9c:			/* STZ aaaa */
    mne = "STZ %04X"; n = 2;
    break;
  case 0x9e:			/* STZ aaaa,X */
    mne = "STZ %04X,X"; n = 2;
    break;
  case 0xaa:			/* TAX */
    mne = "TAX"; n = 0; 
    break;
  case 0xa8:			/* TAY */
    mne = "TAY"; n = 0; 
    break;
  case 0x14:			/* TRB aa */
    mne = "TRB %02X"; n = 1;
    break;
  case 0x1c:			/* TRB aaaa */
    mne = "TRB %04X"; n = 2;
    break;
  case 0x04:			/* TSB aa */
    mne = "TSB %02X"; n = 1;
    break;
  case 0x0c:			/* TSB aaaa */
    mne = "TSB %04X"; n = 2;
    break;
  case 0xba:			/* TSX */
    mne = "TSX"; n = 0; 
    break;
  case 0x8a:			/* TXA */
    mne = "TXA"; n = 0; 
    break;
  case 0x9a:			/* TXS */
    mne = "TXS"; n = 0; 
    break;
  case 0x98:			/* TYA */
    mne = "TYA"; n = 0; 
    break;
  default:			/* Undefined opcode */
    mne = "???"; n = 0; 
    break;
  }

  switch (n)
  {
  case -1:
    sprintf (buf, mne, Pc + (char)READLOC (Pc + 1) + 2);
    *len = 2;
    break;
  case 0:
    sprintf (buf, mne);
    *len = 1;
    break;
  case 1:
    sprintf (buf, mne, READLOC (Pc + 1));
    *len = 2;
    break;
  case 2:
    sprintf (buf, mne, READLOC (Pc + 1) + 256 * READLOC (Pc + 2));
    *len = 3;
    break;
  }

  return buf;
}

- debug:sender
{
  char buf[256];
  int len;

  
  [[self window] makeKeyAndOrderFront:self];

  Debugging = 1;

  mutex_lock (debugLock);
  condition_signal (debugStep);
  condition_wait (debugStepped, debugLock);
  mutex_unlock (debugLock);
  sprintf (buf, "A=%02X X=%02X Y=%02X S=%02X P=%i%i%i%i%i%i%i%i RS=%02X WS=%02X | %04X: %s\n",
	   A, X, Y, Sp, N!=0, V!=0, 0, B!=0, D!=0, I!=0, Z==0, C!=0, ReadState, WriteState, Pc, [self disasm:Pc bytes:&len]);
  [self appendString:buf];
  [self appendString:"bug> "];
  
  return self;
}

- doCommand:(char *) cmd
{
  static char buf[256], buf2[256], da2[256];
  char *da;
  static int lto;
  int dest = 0, from = 0, to = -1;
  int func;
  int pos = 0, opos = strlen (cmd);
  int i, j = 0;
  

  do {
    while (sscanf (&cmd[pos], "%x%n", &to, &opos) > 0) {
      pos += opos;
      switch (cmd[pos]) {
      case '<':
	dest = to;
	pos++;
	break;
      case '.':
	from = to;
	pos++;
	break;
      }
    }
    func = cmd[pos];

    switch (func) {
    case 0:
      if (to < 0) {
	mutex_lock (debugLock);
	condition_signal (debugStep);
	condition_wait (debugStepped, debugLock);
	mutex_unlock (debugLock);
	sprintf (buf, "A=%02X X=%02X Y=%02X S=%02X P=%i%i%i%i%i%i%i%i RS=%02X WS=%02X | %04X: %s\n",
		 A, X, Y, Sp, N!=0, V!=0, 0, B!=0, D!=0, I!=0, Z==0, C!=0, ReadState, WriteState, Pc, [self disasm:Pc bytes:&to]);
	[self appendString:buf];
      }
      else {
	if (from) {
	  for (i = from; i <= to; i++) {
	    if (!((i - from) % 16))
	      sprintf (buf2, "%04X:", i);

	    sprintf (buf, "%02X ", READLOC(i));
	    strcat (buf2, buf);
	    
	    if ((i - from) % 16 == 15) {
	      buf2[strlen (buf2) - 1] = 0;
	      strcat (buf2, "-");
	      for (j = i - 15; j <= i; j++) {
		sprintf (buf, "%c", (READLOC(j) & 0x7f) < 32 ? '.' : READLOC(j) & 0x7f);
		strcat (buf2, buf);
	      }
	      strcat (buf2, "\n");
	      [self appendString:buf2];
	    }
	  }
	  if ((i - from) % 16) {
	    buf2[strlen (buf2) - 1] = 0;
	    strcat (buf2, "-");
	    for (j = i - (i - from) % 16; j < i; j++) {
	      sprintf (buf, "%c", (READLOC(j) & 0x7f) < 32 ? '.' : READLOC(j) & 0x7f);
	      strcat (buf2, buf);
	    }
	    strcat (buf2, "\n");
	    [self appendString:buf2];
	  }
	}
	else {
	  sprintf (buf, "%04X:%02X-%c\n", to, READLOC(to), (READLOC(to) & 0x7f) < 32 ? '.' : READLOC(to) & 0x7f);
	  [self appendString:buf];
	}
      }
      break;

    case 'w':
      if (from) {
	for (i = from; i <= to; i += 2) {
	  if (!((i - from) % 16))
	    sprintf (buf2, "%04X:", i);

	  sprintf (buf, "%04X ", READLOC(i) | (READLOC(i+1) << 8));
	  strcat (buf2, buf);
	    
	  if (((i + 1) - from) % 16 == 15) {
	    buf2[strlen (buf2) - 1] = 0;
	    strcat (buf2, "-");
	    for (j = i - 14; j <= i + 1; j++) {
	      sprintf (buf, "%c", (READLOC(j) & 0x7f) < 32 ? '.' : READLOC(j) & 0x7f);
	      strcat (buf2, buf);
	    }
	    strcat (buf2, "\n");
	    [self appendString:buf2];
	  }
	}
	if (((i + 1) - from) % 16) {
	  buf2[strlen (buf2) - 1] = 0;
	  strcat (buf2, "-");
	  for (j = (i + 1) - ((i + 1) - from) % 16; j <= i; j++) {
	    sprintf (buf, "%c", (READLOC(j) & 0x7f) < 32 ? '.' : READLOC(j) & 0x7f);
	    strcat (buf2, buf);
	  }
	  strcat (buf2, "\n");
	  [self appendString:buf2];
	}
      }
	else {
	  sprintf (buf, "%04X:%04X-%c%c\n", to, READLOC(to) | (READLOC(to+1) << 8),
		   (READLOC(to) & 0x7f) < 32 ? '.' : READLOC(to) & 0x7f,
		   (READLOC(to+1) & 0x7f) < 32 ? '.' : READLOC(to+1) & 0x7f);
	  [self appendString:buf];
	}
      break;

    case 'r':
    case 'R':
      Debugging = 0;
      mutex_lock (debugLock);
      condition_signal (debugStep);
      mutex_unlock (debugLock);
      break;

    case 'l':
    case 'L':
      if (to < 0)
	to = lto;
      /* Disassemble Prodos MLI calls just like the GS can */
      for (i = 0; i < 20; i++, to += from) {
	if (READLOC (to - 3) == 0x20 && READLOC (to - 2) == 0x00 && READLOC (to - 1) == 0xBF) {
	  sprintf (da2, "    %02X", READLOC (to));
	  da = da2;
	  from = 1;
	}
	else if (READLOC (to - 4) == 0x20 && READLOC (to - 3) == 0x00 && READLOC (to - 2) == 0xBF) {
	  sprintf (da2, "    %04X", READLOC (to) + READLOC (to + 1) * 0x100);
	  da = da2;
	  from = 2;
	}
	else
	  da = [self disasm:to bytes:&from];
	sprintf (buf2, "%04X-   ", to);
	for (j = 0; j < from; j++) {
	  sprintf (buf, "%02X ", READLOC(to + j));
	  strcat (buf2, buf);
	}
	for (dest = j; dest < 4; dest++)
	  strcat (buf2, "   ");
	strcat (buf2, " ");
	strcat (buf2, da);
	strcat (buf2, "\n");
	[self appendString:buf2];
      }
      lto = to;
      break;

    case '=':
      if (to < 0) {
	sprintf (buf, "Hex-> $%08X\n", atoi (&cmd[pos+1]));
	pos = strlen (cmd);
	[self appendString:buf];
      }
      else {
	pos++;
	switch (cmd[pos]) {
	case 'A':
	  A = to;
	  break;
	case 'X':
	  X = to;
	  break;
	case 'Y':
	  Y = to;
	  break;
	case 'S':
	  Sp = to;
	  break;
	case 'C':
	  Pc = to;
	  break;
	case 'B':
	  BreakPoint = to;
	  break;
	case 0:
	  pos--;
	  sprintf (buf, "Decimal-> %i\n", to);
	  [self appendString:buf];
	}
      }
      break;

    case '?':
      sprintf (buf, "A=%02X X=%02X Y=%02X S=%02X P=%i%i%i%i%i%i%i%i RS=%02X WS=%02X PC=%04X\n",
	       A, X, Y, Sp, N!=0, V!=0, 0, B!=0, D!=0, I!=0, Z==0, C!=0, ReadState, WriteState, Pc);
      [self appendString:buf];
      break;

    default:
      break;
    }

    pos++;
  } while (cmd[pos]);

  if (Debugging)
    [self appendString:"bug> "];
  
  return self;
}

- initFrame:(const NXRect *)frameRect
{	
  [super initFrame:frameRect];
  [self setVertScrollerRequired: YES];
  [self setBackgroundGray: NX_WHITE];
  [self awake];

  return self;
}

- awake
{    
  NXRect textRect;
  id theText;
    
  textRect.origin.x = textRect.origin.y = 0.0;
  [self getContentSize: &(textRect.size)];
  theText =
    [[Text alloc]
   initFrame: &textRect  text:NULL alignment: NX_LEFTALIGNED];
  [theText notifyAncestorWhenFrameChanged:YES];
  [theText setHorizResizable:NO];
  [theText setVertResizable:YES];
    
  textRect.size.width = 0.0;
  [theText setMinSize:&(textRect.size)];
  [self getContentSize: &(textRect.size)];
  textRect.size.height = 1000000;
  [theText setMaxSize:&(textRect.size)];
    
  [[theText superview] setAutoresizeSubviews:YES];
  [[theText superview] setAutosizing: NX_HEIGHTSIZABLE | NX_WIDTHSIZABLE];
    
  [theText setCharFilter: NXFieldFilter];
  [theText setMonoFont:TRUE];
  [theText setFont:[Font newFont:"Ohlfs" size:10]];
  [self setDocView: theText];

  return self;
}

- setDocView:anObject
{
  [super setDocView:anObject];
  docView = anObject;
  [docView setDelegate:self];
  return self;
}

- appendString:(char *)buffer
    // append the buffer to the end of the text object
{
  [docView setSel:[docView textLength] :0];
  [docView replaceSel:buffer];
  [docView scrollSelToVisible];
  lastTextCount = [docView textLength];
  
  return self;
}

/* Text delegate methods*/
- (BOOL)textWillChange:textObject 
    // moving the selection to the end before the change means the keys
    // the user hits will always appear at the end of the scroll view
{
  [docView setSel:[docView textLength] :0];
  [docView scrollSelToVisible];
    
  return NO;
}

- textDidEnd:textObject endChar:(unsigned short)whyEnd  
    // Get this on every Return if using NXFieldFilter
    // really simplistic.  Assumes user does not type ahead
{
  char buf[256];
  int len;


  len = [docView textLength] - lastTextCount;
  [docView getSubstring:buf start:([docView textLength]-len) length:len];
  buf[len] = 0;
  
  [docView setSel:[docView textLength] :0];
  [docView replaceSel: "\n"];
  lastTextCount = [docView textLength];

  [self doCommand:buf];
  
  return self;
}

@end
