/*
 * VideoView.m - handles conversion of apple screens
 * 		 and fetching of keys
 * written by Chris Osborn <osborn@athena.ecs.csus.edu>
 * Hi-Res routines written by Mathew J. Hostetter <mjhostet@slayer.MIT.EDU>
 */

#import "VideoView.h"
#import "apple.h"
#import "hgrconv.h"
#import "hgrtable.h"
#import <appkit/graphics.h>
#import <appkit/Application.h>
#import <appkit/NXBitmapImageRep.h>
#import <appkit/NXImage.h>
#import <appkit/tiff.h>
#import <string.h>
#import <dpsclient/psops.h>
#import <dpsclient/wraps.h>
#import <dpsclient/event.h>
#import <dpsclient/dpsNeXT.h>
#import <cthreads.h>


#define MODE_LORES 0x100
#define MODE_HIRES 0x200
#define MODE_80COL 0x400
#define MODE_MOUSE 0x800

#define VModes(_x,_y) (MODE_LORES * (GraphicsOn && !HiResOn && !(_y >= 20 && MixedOn)) | \
		  MODE_HIRES * (GraphicsOn && HiResOn && !(_y >= 20 && MixedOn)) | \
		  MODE_80COL * (Vid80On) | \
		  MODE_MOUSE * (AltChar))
#define Visible(x,y) (!(GraphicsOn && HiResOn && !(y >= 20 && MixedOn)))

int WHiResOn = 0;
id videoScreen;
extern int LBasCalc[];
extern int LHRBasCalc[];
int doScroll;
int bufferKeys;
int keyboardBuffer[32];
int keyboardPos = 0;
mutex_t videoUpdating;

@implementation VideoView

- initdisplay
{
  int i;

  
  if (!(col40images = [NXImage findImageNamed:"col40"]))
    fprintf (stderr, "Can't load col40.tiff\n");
  if (!(col80images = [NXImage findImageNamed:"col80"]))
    fprintf (stderr, "Can't load col80.tiff\n");

  for (i = 0; i < 1920; i++)
    otscreen[i] = wuzthere[i] = 0xff;

  for (i = 0; i < 0x2000; i++)
    lhscreen[i] = i & 0xff;

  videoUpdating = mutex_alloc ();
  
  videoScreen = self;
  DPSSetDeadKeysEnabled ([NXApp context], NO);

  [[self window] makeFirstResponder:self];
  [self refresh];
  [[self window] makeKeyAndOrderFront:self];
  NXPing ();

  return self;
}

- initFrame:(const NXRect *) frameRect
{
  [super initFrame:frameRect];

  memset (fbuf, 0, sizeof fbuf);            /* Clear to black. */
  return self;
}

- drawSelf:(const NXRect *) rects: (int) rectCount
{
  NXImageBitmap (&bounds, 280, 192, 2, 1, NX_MESHED, NX_MONOTONICMASK, fbuf,
		 NULL, NULL, NULL, NULL);

  return self;
}


- drawChangedAreas:(const unsigned char *) a2_screen shadow:(unsigned char *) a2_shadow base:(int) bot
{
  unsigned char changed[12];
  NXRect r = bounds;
  int i, start, height;


  [self lockFocus];

  /* Find out what areas changed, if any. */
  if (find_changed_areas (a2_screen, a2_shadow, changed) || !WHiResOn) {
    for (i = 0; i < bot; i++)
      if (changed[i] || !WHiResOn) {
	start = i * 16;
	for (i++; i < bot && (changed[i] || !WHiResOn); i++);
	height = i * 16 - start;
	r.origin.y = (192 - (start + height)) * 2.0;
	r.size.height = height * 2.0; /*  * 2.0 since we are scaled. */
	apple_to_next (a2_screen, fbuf, start, height);
#if USE_COLOR
	r.size.width = 288 * 2;
	NXImageBitmap (&r, 288, height, 2, 1, NX_MESHED, NX_MONOTONICMASK,
		       &fbuf[start][0], NULL, NULL, NULL, NULL);
#else	    
	NXImageBitmap (&r, 280, height, 2, 1, NX_MESHED, NX_MONOTONICMASK,
		       &fbuf[start][0], NULL, NULL, NULL, NULL);
#endif
      }
    copy_changed_areas (a2_screen, a2_shadow, changed);
  }

  [self unlockFocus];

  return self;
}

- displayChar:(int) c at:(int) x :(int) y
{
  int ax, au;
  int d = c | VModes(x,y);
  NXRect r;
  NXPoint p;


  if (Vid80On) {
    if (Visible (x, y) && wuzthere[y * 80 + x] != d) {
      au = x % 2;
      ax = x / 2;
      r.origin.x = (c % 32) * 7 + 224 * ((d & MODE_MOUSE) > 0);
      r.origin.y = 112 - (c / 32) * 16 + ((d & MODE_LORES) / 2);
      r.size.width = 7;
      r.size.height = 16;
      p.x = ax * 14 + 7 * au;
      p.y = (23 - y) * 16;
      [col80images composite:NX_COPY fromRect:&r toPoint:&p];
    }
    wuzthere[y * 80 + x] = d;
  }
  else {
    if (Visible (x, y) && wuzthere[y * 80 + x * 2] != d) {
      r.origin.x = (c % 32) * 14 + 448 * ((d & MODE_MOUSE) > 0);
      r.origin.y = 112 - (c / 32) * 16 + ((d & MODE_LORES) / 2);
      r.size.width = 14;
      r.size.height = 16;
      p.x = x * 14;
      p.y = (23 - y) * 16;
      [col40images composite:NX_COPY fromRect:&r toPoint:&p];
    }
    wuzthere[y * 80 + x * 2] = wuzthere[y * 80 + x * 2 + 1] = d;
  }
  
  return self;
}
  
- refresh
{
  int x, y;
  extern int VideoModeChange;


  mutex_lock (videoUpdating);
  
  switch (doScroll) {
  case 1:
    [self scroll40];
    break;
  case 2:
    [self scroll80];
    break;
  }
  
  if (GraphicsOn && HiResOn && !Store80) /* What does a real //e do? */
    [self drawChangedAreas:&SysMem[0x2000 + 0x2000 * ShowPage] shadow:lhscreen base:10 + 2 * (!MixedOn)];

  bcopy (&SysMem[0x400], tscreen, 0x400);
  bcopy (&SysMem[0x400 + AUX48], &tscreen[0x400], 0x400);
  if (VideoModeChange || bcmp (tscreen, otscreen, 0x400) || (Vid80On && bcmp (&tscreen[0x400], &otscreen[0x400], 0x400)) || (GraphicsOn && HiResOn)) {
    [self lockFocus];  
    if (Vid80On) {
      for (x = 0; x < 80; x++)
	for (y = 0; y < 24; y++)
	  [self displayChar:tscreen[LBasCalc[y] + (x / 2) + (1 - x % 2) * 0x400 - 0x400] at:x :y];
    }
    else {
      for (x = 0; x < 40; x++)
	for (y = 0; y < 24; y++)
	  [self displayChar:tscreen[LBasCalc[y] + x - 0x400] at:x :y];
    }
    PSflushgraphics ();
    [self unlockFocus];
    bcopy (tscreen, otscreen, 0x800);
  }
  VideoModeChange = 0;
  
  WHiResOn = GraphicsOn && HiResOn;

  mutex_unlock (videoUpdating);
  
  return self;
}

- scroll40
{
  int i, j;


  [self lockFocus];

  i = SysMem[34];
  if (GraphicsOn && HiResOn) {
    i = 20;
    if (!MixedOn)
      i = 24;
  }
  if (SysMem[33] && SysMem[35] > i) {
    PScomposite (SysMem[32] * 14, (24 - SysMem[35]) * 16, SysMem[33] * 14, (SysMem[35] - i - 1) * 16, NXNullObject,
		 SysMem[32] * 14, (24 - SysMem[35] + 1) * 16, NX_COPY);
    if (SysMem[50] & 0x80) {
      PSsetgray (0);
      PScompositerect (SysMem[32] * 14, (24 - SysMem[35]) * 16, SysMem[33] * 14, 16, NX_COPY);
    }
  }
  
  for (i = SysMem[34] + 1; i < SysMem[35]; i++)
    for (j = SysMem[32]; j < SysMem[33] + SysMem[32]; j++)
      wuzthere[(i - 1) * 80 + j * 2] = wuzthere[(i - 1) * 80 + j * 2 + 1] = wuzthere[i * 80 + j * 2];

  if (SysMem[50] & 0x80)
    for (j = SysMem[32]; j < SysMem[33] + SysMem[32]; j++)
      wuzthere[(i - 1) * 80 + j * 2] = wuzthere[(i - 1) * 80 + j * 2 + 1] = (0xa0 & SysMem[50]) | VModes(j,i);

  [self unlockFocus];

  doScroll = 0;
  
  return self;
}

- scroll80
{
  int i, j;


  [self lockFocus];

  i = SysMem[34];
  if (GraphicsOn && HiResOn) {
    i = 20;
    if (!MixedOn)
      i = 24;
  }
  if (SysMem[33] && SysMem[35] > i) {
    PScomposite (SysMem[32] * 7, (24 - SysMem[35]) * 16, SysMem[33] * 7, (SysMem[35] - i - 1) * 16, NXNullObject,
		 SysMem[32] * 7, (24 - SysMem[35] + 1) * 16, NX_COPY);
    if (SysMem[50] & 0x80) {
      PSsetgray (0);
      PScompositerect (SysMem[32] * 7, (24 - SysMem[35]) * 16, SysMem[33] * 7, 16, NX_COPY);
    }
  }
  
  for (i = SysMem[34] + 1; i < SysMem[35]; i++)
    for (j = SysMem[32]; j < SysMem[33] + SysMem[32]; j++)
      wuzthere[(i - 1) * 80 + j] = wuzthere[i * 80 + j];


  if (SysMem[50] & 0x80)
    for (j = SysMem[32]; j < SysMem[33] + SysMem[32]; j++)
      wuzthere[(i - 1) * 80 + j] = 0xa0 | VModes(j,i);

  [self unlockFocus];

  doScroll = 0;
  
  return self;
}

- keyDown:(NXEvent *) theEvent
{
  int key = theEvent->data.key.charCode;


  if (theEvent->flags & NX_NUMERICPADMASK)
    switch (theEvent->data.key.charCode) {
    case 0xAC: /* Left Arrow */
      key = 0x08;
      break;
    case 0xAD: /* Up Arrow */
      key = 0x0B;
      break;
    case 0xAE: /* Right Arrow */
      key = 0x15;
      break;
    case 0xAF: /* Down Arrow */
      key = 0x0A;
      break;
    }

  if (bufferKeys) {
    if ((key == 127) && AppleKey && (theEvent->flags & NX_CONTROLMASK)) {
      /* Clear buffer. Apple-Control-Delete, just like GS.
	 Sure wish the NeXT had a way to do it in other apps. */
      keyboardPos = 0;
      MegaLastKey = 0;
      return self;
    }
    if (keyboardPos < 32) {
      keyboardBuffer[keyboardPos] = key;
      keyboardPos++;
    }
  }
  else
    MegaLastKey = key | 0x80;

  SysMem[0xc010] = 0x80;
  
  PSobscurecursor ();
  
  return self;
}

- keyUp:(NXEvent *) theEvent
{
  SysMem[0xc010] = 0;
  
  return self;
}

-(BOOL) acceptsFirstResponder
{
  return YES;
}

@end

void videoUpdate ()
{
  [videoScreen refresh];

  return;
}
