/*
 * mem.c -- Deals with setup and state of memory
 * written by Chris Osborn <osborn@athena.ecs.csus.edu>
 */

#include "apple.h"

BYTE SysMem[0x28000];  /* Memory and rom for entire machine */

int AltZP       = 0;  /* set if Alt ZP switched in     */
int AuxRead     = 0;  /* set if Aux 48k readable       */
int AuxWrite    = 0;  /* set if Aux 48k writeable      */
int LCRead      = 0;  /* set if 16K RAM readable       */
int LCWrite     = 0;  /* set if 16K RAM writeable      */
int Bank2Enable = 0;  /* set if bank 2 Ram enabled     */
int Store80     = 0;  /* set if 80 column store        */
int ShowPage    = 0;  /* set if secondary page visible */
int HiResOn     = 0;  /* set if hires visible          */
int SlotCXRom   = 0;  /* set if use internal rom       */
int SlotC3Rom   = 0;  /* set if use card rom           */

int ReadState = 0;
int WriteState = 0;

int bankreadoffset[7][128] = { /* table that contains offset into SysMem */
  { /* Location of zero page (0x0000 - 0x01ff) */
    MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48,
    MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48,
    MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48,
    MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48,
    MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48,
    MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48,
    MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48,
    MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48,
    MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48,
    MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48,
    MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48,
    MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48,
    MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48, 
  },
  { /* Location of Page1 text screen (0x0400 - 0x07ff) */
    MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48,
    AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48, MAIN48, MAIN48,
    MAIN48, MAIN48, MAIN48, MAIN48, MAIN48, MAIN48, MAIN48, MAIN48, MAIN48,
    MAIN48, MAIN48, MAIN48, MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48, AUX48,
    AUX48, MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48, AUX48,
    AUX48, AUX48, AUX48, AUX48, AUX48, AUX48, AUX48, AUX48, AUX48, AUX48, AUX48,
    AUX48, AUX48, AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48,
    AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48,
    MAIN48, MAIN48, MAIN48, MAIN48, MAIN48, MAIN48, MAIN48, MAIN48, MAIN48,
    MAIN48, MAIN48, MAIN48, MAIN48, MAIN48, MAIN48, MAIN48, MAIN48, MAIN48,
    AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48,
    MAIN48, MAIN48, AUX48, AUX48, AUX48, AUX48, AUX48, AUX48, AUX48, AUX48,
    AUX48, AUX48, AUX48, AUX48, AUX48, AUX48, AUX48, AUX48, AUX48, AUX48,
  },
  { /* Location of Page1 Hires screen (0x2000 - 0x3fff) */
    MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48,
    AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48,
    MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48,
    AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48,
    MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48,
    AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48,
    MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48,
    AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48,
    MAIN48, MAIN48, MAIN48, MAIN48, MAIN48, MAIN48, MAIN48, MAIN48, MAIN48,
    MAIN48, MAIN48, MAIN48, MAIN48, MAIN48, MAIN48, MAIN48, MAIN48, MAIN48,
    AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48,
    MAIN48, MAIN48, AUX48, AUX48, AUX48, AUX48, AUX48, AUX48, AUX48, AUX48,
    AUX48, AUX48, AUX48, AUX48, AUX48, AUX48, AUX48, AUX48, AUX48, AUX48,
  },
  { /* Location of normal 48k (0x0200 - 0x03ff, 0x0800 - 0x1fff, 0x4000 - 0xbfff) */
    MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48,
    AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48,
    MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48,
    AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48,
    MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48,
    AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48,
    MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48,
    AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48,
    MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48,
    AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48,
    MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48,
    AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48,
    MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48,
  },
  { /* Location of LC bank (0xd000 - 0xdfff) */
    ROM, ROM, ROM, ROM, LCBANK1, ALCBANK1, LCBANK1, ALCBANK1, ROM, ROM, ROM,
    ROM, LCBANK2, ALCBANK2, LCBANK2, ALCBANK2, ROM, ROM, ROM, ROM, LCBANK1,
    ALCBANK1, LCBANK1, ALCBANK1, ROM, ROM, ROM, ROM, LCBANK2, ALCBANK2, LCBANK2,
    ALCBANK2, ROM, ROM, ROM, ROM, LCBANK1, ALCBANK1, LCBANK1, ALCBANK1, ROM,
    ROM, ROM, ROM, LCBANK2, ALCBANK2, LCBANK2, ALCBANK2, ROM, ROM, ROM, ROM,
    LCBANK1, ALCBANK1, LCBANK1, ALCBANK1, ROM, ROM, ROM, ROM, LCBANK2, ALCBANK2,
    LCBANK2, ALCBANK2, ROM, ROM, ROM, ROM, LCBANK1, ALCBANK1, LCBANK1, ALCBANK1,
    ROM, ROM, ROM, ROM, LCBANK2, ALCBANK2, LCBANK2, ALCBANK2, ROM, ROM, ROM,
    ROM, LCBANK1, ALCBANK1, LCBANK1, ALCBANK1, ROM, ROM, ROM, ROM, LCBANK2,
    ALCBANK2, LCBANK2, ALCBANK2, ROM, ROM, ROM, ROM, LCBANK1, ALCBANK1, LCBANK1,
    ALCBANK1, ROM, ROM, ROM, ROM, LCBANK2, ALCBANK2, LCBANK2, ALCBANK2, ROM,
    ROM, ROM, ROM, LCBANK1, ALCBANK1, LCBANK1, ALCBANK1, ROM, ROM, ROM, ROM,
    LCBANK2, ALCBANK2, LCBANK2, ALCBANK2,
  },
  { /* Location of Rom or LC (0xe000 - 0xffff) */
    ROM, ROM, ROM, ROM, MAIN48, AUX48, MAIN48, AUX48, ROM, ROM, ROM, ROM,
    MAIN48, AUX48, MAIN48, AUX48, ROM, ROM, ROM, ROM, MAIN48, AUX48, MAIN48,
    AUX48, ROM, ROM, ROM, ROM, MAIN48, AUX48, MAIN48, AUX48, ROM, ROM, ROM, ROM,
    MAIN48, AUX48, MAIN48, AUX48, ROM, ROM, ROM, ROM, MAIN48, AUX48, MAIN48,
    AUX48, ROM, ROM, ROM, ROM, MAIN48, AUX48, MAIN48, AUX48, ROM, ROM, ROM, ROM,
    MAIN48, AUX48, MAIN48, AUX48, ROM, ROM, ROM, ROM, MAIN48, AUX48, MAIN48,
    AUX48, ROM, ROM, ROM, ROM, MAIN48, AUX48, MAIN48, AUX48, ROM, ROM, ROM, ROM,
    MAIN48, AUX48, MAIN48, AUX48, ROM, ROM, ROM, ROM, MAIN48, AUX48, MAIN48,
    AUX48, ROM, ROM, ROM, ROM, MAIN48, AUX48, MAIN48, AUX48, ROM, ROM, ROM, ROM,
    MAIN48, AUX48, MAIN48, AUX48, ROM, ROM, ROM, ROM, MAIN48, AUX48, MAIN48,
    AUX48, ROM, ROM, ROM, ROM, MAIN48, AUX48, MAIN48, AUX48,
  },
  { /* Location of ROM (0xc100 - 0xcfff) */
    ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, 
    ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, 
    ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, 
    ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, 
    ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, 
    ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, 
    ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, 
    ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM, ROM,
  }
};

int bankwriteoffset[7][128] = { /* table that contains offset into SysMem */
  { /* Location of zero page (0x0000 - 0x01ff) */
    MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48,
    MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48,
    MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48,
    MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48,
    MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48,
    MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48,
    MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48,
    MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48,
    MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48,
    MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48,
    MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48,
    MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48,
    MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48, MAIN48, AUX48, 
  },
  { /* Location of Page1 text screen (0x0400 - 0x07ff) */
    MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48,
    AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48, MAIN48, MAIN48,
    MAIN48, MAIN48, MAIN48, MAIN48, MAIN48, MAIN48, MAIN48, MAIN48, MAIN48,
    MAIN48, MAIN48, MAIN48, MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48, AUX48,
    AUX48, MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48, AUX48,
    AUX48, AUX48, AUX48, AUX48, AUX48, AUX48, AUX48, AUX48, AUX48, AUX48, AUX48,
    AUX48, AUX48, AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48,
    AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48,
    MAIN48, MAIN48, MAIN48, MAIN48, MAIN48, MAIN48, MAIN48, MAIN48, MAIN48,
    MAIN48, MAIN48, MAIN48, MAIN48, MAIN48, MAIN48, MAIN48, MAIN48, MAIN48,
    AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48,
    MAIN48, MAIN48, AUX48, AUX48, AUX48, AUX48, AUX48, AUX48, AUX48, AUX48,
    AUX48, AUX48, AUX48, AUX48, AUX48, AUX48, AUX48, AUX48, AUX48, AUX48,
  },
  { /* Location of Page1 Hires screen (0x2000 - 0x3fff) */
    MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48,
    AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48,
    MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48,
    AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48,
    MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48,
    AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48,
    MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48,
    AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48,
    MAIN48, MAIN48, MAIN48, MAIN48, MAIN48, MAIN48, MAIN48, MAIN48, MAIN48,
    MAIN48, MAIN48, MAIN48, MAIN48, MAIN48, MAIN48, MAIN48, MAIN48, MAIN48,
    AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48,
    MAIN48, MAIN48, AUX48, AUX48, AUX48, AUX48, AUX48, AUX48, AUX48, AUX48,
    AUX48, AUX48, AUX48, AUX48, AUX48, AUX48, AUX48, AUX48, AUX48, AUX48,
  },
  { /* Location of normal 48k (0x0200 - 0x03ff, 0x0800 - 0x1fff, 0x4000 - 0xbfff) */
    MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48,
    AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48,
    MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48,
    AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48,
    MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48,
    AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48,
    MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48,
    AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48,
    MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48,
    AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48,
    MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48,
    AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48,
    MAIN48, MAIN48, AUX48, AUX48, MAIN48, MAIN48, AUX48, AUX48,
  },
  { /* Location of LC bank (0xd000 - 0xdfff) */
    JUNK, JUNK, JUNK, JUNK, LCBANK1, ALCBANK1, LCBANK1, ALCBANK1, JUNK, JUNK, JUNK,
    JUNK, LCBANK2, ALCBANK2, LCBANK2, ALCBANK2, JUNK, JUNK, JUNK, JUNK, LCBANK1,
    ALCBANK1, LCBANK1, ALCBANK1, JUNK, JUNK, JUNK, JUNK, LCBANK2, ALCBANK2, LCBANK2,
    ALCBANK2, JUNK, JUNK, JUNK, JUNK, LCBANK1, ALCBANK1, LCBANK1, ALCBANK1, JUNK,
    JUNK, JUNK, JUNK, LCBANK2, ALCBANK2, LCBANK2, ALCBANK2, JUNK, JUNK, JUNK, JUNK,
    LCBANK1, ALCBANK1, LCBANK1, ALCBANK1, JUNK, JUNK, JUNK, JUNK, LCBANK2, ALCBANK2,
    LCBANK2, ALCBANK2, JUNK, JUNK, JUNK, JUNK, LCBANK1, ALCBANK1, LCBANK1, ALCBANK1,
    JUNK, JUNK, JUNK, JUNK, LCBANK2, ALCBANK2, LCBANK2, ALCBANK2, JUNK, JUNK, JUNK,
    JUNK, LCBANK1, ALCBANK1, LCBANK1, ALCBANK1, JUNK, JUNK, JUNK, JUNK, LCBANK2,
    ALCBANK2, LCBANK2, ALCBANK2, JUNK, JUNK, JUNK, JUNK, LCBANK1, ALCBANK1, LCBANK1,
    ALCBANK1, JUNK, JUNK, JUNK, JUNK, LCBANK2, ALCBANK2, LCBANK2, ALCBANK2, JUNK,
    JUNK, JUNK, JUNK, LCBANK1, ALCBANK1, LCBANK1, ALCBANK1, JUNK, JUNK, JUNK, JUNK,
    LCBANK2, ALCBANK2, LCBANK2, ALCBANK2,
  },
  { /* Location of JUNK or LC (0xe000 - 0xffff) */
    JUNK, JUNK, JUNK, JUNK, MAIN48, AUX48, MAIN48, AUX48, JUNK, JUNK, JUNK, JUNK,
    MAIN48, AUX48, MAIN48, AUX48, JUNK, JUNK, JUNK, JUNK, MAIN48, AUX48, MAIN48,
    AUX48, JUNK, JUNK, JUNK, JUNK, MAIN48, AUX48, MAIN48, AUX48, JUNK, JUNK, JUNK, JUNK,
    MAIN48, AUX48, MAIN48, AUX48, JUNK, JUNK, JUNK, JUNK, MAIN48, AUX48, MAIN48,
    AUX48, JUNK, JUNK, JUNK, JUNK, MAIN48, AUX48, MAIN48, AUX48, JUNK, JUNK, JUNK, JUNK,
    MAIN48, AUX48, MAIN48, AUX48, JUNK, JUNK, JUNK, JUNK, MAIN48, AUX48, MAIN48,
    AUX48, JUNK, JUNK, JUNK, JUNK, MAIN48, AUX48, MAIN48, AUX48, JUNK, JUNK, JUNK, JUNK,
    MAIN48, AUX48, MAIN48, AUX48, JUNK, JUNK, JUNK, JUNK, MAIN48, AUX48, MAIN48,
    AUX48, JUNK, JUNK, JUNK, JUNK, MAIN48, AUX48, MAIN48, AUX48, JUNK, JUNK, JUNK, JUNK,
    MAIN48, AUX48, MAIN48, AUX48, JUNK, JUNK, JUNK, JUNK, MAIN48, AUX48, MAIN48,
    AUX48, JUNK, JUNK, JUNK, JUNK, MAIN48, AUX48, MAIN48, AUX48,
  },
  { /* Location of JUNK (0xc100 - 0xcfff) */
    JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, 
    JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, 
    JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, 
    JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, 
    JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, 
    JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, 
    JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, 
    JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK, JUNK,
  }
};

int bankselect[256] = {
  0, 0, 3, 3, 1, 1, 1, 1, 3, 3, 3, 3, 3, 3, 3, 3,
  3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
  2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
  2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
  3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
  3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
  3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
  3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
  3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
  3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
  3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
  3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
 -1, 6, 6,14, 6, 6, 6, 6,14,14,14,14,14,14,14,14,
  4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
};

void changestates ()
{
  ReadState = AltZP | AuxRead << 1 | LCRead << 2 | Bank2Enable << 3 | Store80 << 4 | ShowPage << 5 | HiResOn << 6;
  WriteState = AltZP | AuxWrite << 1 | LCWrite << 2 | Bank2Enable << 3 | Store80 << 4 | ShowPage << 5 | HiResOn << 6;
  
  return;
}
