/*
 * unixShare.c - this handles mapping of the UNIX
 * 		 filesystem into a ProDOS compatible format
 * written by Chris Osborn <osborn@athena.ecs.csus.edu>
 */

#import <stdio.h>
#import <sys/param.h>
#import <strings.h>
#import <sys/types.h>
#import <sys/stat.h>
#import <sys/errno.h>
#import <sys/dir.h>
#import "apple.h"

extern long ptime (time_t t);
extern int getuid ();
extern int getgid ();

#define UNIX_COMMAND	0xc796
#define UNIX_PARM	0xc797

#define UNIX_DIRECTORY 0x0100

#define UNIX_SLOTDRIVE 0x20

#define READWORD(v) ({int _a1 = v; READLOC(_a1) | (READLOC(_a1+1) << 8);})

#define UPPER(a) ({int _a1 = a; _a1 >= 'a' && _a1 <= 'z' ? _a1 - 32 : _a1;})
#define LOWER(a) ({int _a1 = a; _a1 >= 'A' && _a1 <= 'Z' ? _a1 + 32 : _a1;})

#define LDA(v) (void)({A=v;SIGN(A);ZERO(A);})

#define PDOSERR_NOERROR		0x00
#define PDOSERR_INVALIDMLICMD	0x01
#define PDOSERR_INVALIDPARMCNT	0x04
#define PDOSERR_INTRPTTBLFULL	0x25
#define PDOSERR_IOERROR		0x27
#define PDOSERR_NODEVICECNCTD	0x28
#define PDOSERR_WRITEPROTECTED	0x2b
#define PDOSERR_VOLUMESWITCHED	0x2e
#define PDOSERR_INVALIDPATHSYN	0x40
#define PDOSERR_TOOMANYOPEN	0x42
#define PDOSERR_INVALIDREFNUM	0x43
#define PDOSERR_NOTADIR		0x44
#define PDOSERR_VOLUMENTMNTD	0x45
#define PDOSERR_FILENOTFOUND	0x46
#define PDOSERR_DUPFILENAME	0x47
#define PDOSERR_DISKFULL	0x48
#define PDOSERR_VOLDIRFULL	0x49
#define PDOSERR_INCOMPFILETYPE	0x4a
#define PDOSERR_UNSUPSTORTYPE	0x4b
#define PDOSERR_ENDOFFILE	0x4c
#define PDOSERR_POSPASTEOF	0x4d
#define PDOSERR_ACCESSERROR	0x4e
#define PDOSERR_FILEALRDYOPEN	0x50
#define PDOSERR_FILECNTBAD	0x51
#define PDOSERR_NONPRODOS	0x52
#define PDOSERR_BADPARM		0x53
#define PDOSERR_VCBOVERFLOW	0x55
#define PDOSERR_BADBUFADDR	0x56
#define PDOSERR_DUPVOLUME	0x57
#define PDOSERR_BADVOLBITMAP	0x5a

typedef struct fileInfoRec {
  int accessBits;
  int fileType;
  int auxType;
  int storageType;
  int blocksUsed;
  int fileLength;
  int dateLastMod;
  int timeLastMod;
  int dateCreate;
  int timeCreate;
} fileInfo;
  
char unixPrefix[MAXPATHLEN] = "";
char unixVolumeName[MAXPATHLEN]; /* Name of directory to be volume */
char unixVolumePath[MAXPATHLEN]; /* Path to directory */

FILE *unixFile[8];
int unixFileMode[8];
int unixFilePos[8];
int unixFileLength[8];
int unixFileNewlineMask[8];
int unixFileNewlineChar[8];
int lastCmd = 0;
int oldAddress;

int pdoserrno;

int permissions (int mode, int owner, int group)
{
  int uid = getuid ();
  int gid = getgid ();

  
  if (uid == owner)
    return ((mode & 0700) / 0100);
  if (gid == group)
    return ((mode & 0070) / 0010);
  
  return ((mode & 0007) / 0001);
}

fileInfo *getAppleDoubleInfo (char *filename)
{
  static fileInfo info;
  char path[MAXPATHLEN];
  char file[MAXPATHLEN];
  char buf[MAXPATHLEN];
  struct stat sbuf;
  FILE *ainfo;
  char *p;
  int i;
  
  int mkdir ();
  

  strcpy (path, filename);
  if (p = rindex (path, '/')) {
    strcpy (file, p+1);
    *p = 0;
  }
  else {
    path[0] = 0;
    strcpy (file, filename);
  }

  sprintf (buf, "%s/%s", path, file);
  if (stat (buf, &sbuf)) {
    return NULL;
  }
  else {
    i = permissions (sbuf.st_mode, sbuf.st_uid, sbuf.st_gid);
    info.accessBits =
      ((i & 2) << 6) | /* DESTROY. Wrong. Should check directory it's in */
      ((i & 2) << 5) | /* RENAME. Same as above */
      0x20           | /* BACKUP. No equivalent. Just set it. Maybe put in AD info */
      ((i & 2) << 0) | /* WRITE */
      ((i & 4) >> 2);  /* READ */
    
    info.storageType = 0x01; /* Seedling. What does AppleShare do? */
    info.blocksUsed = sbuf.st_size / 512 + (sbuf.st_size % 512 > 0);
    info.fileLength = sbuf.st_size;
    info.dateLastMod = (ptime (sbuf.st_mtime) >> 16) & 0xffff;
    info.timeLastMod = ptime (sbuf.st_mtime) & 0xffff;
    info.dateCreate = (ptime (sbuf.st_ctime) >> 16) & 0xffff;
    info.timeCreate = ptime (sbuf.st_ctime) & 0xffff;
  }

  if (sbuf.st_mode & S_IFDIR) {
    info.storageType = 0x0d;
    info.fileType = 0x0f;
    info.auxType = 0x0000;
  }
  else {
    sprintf (buf, "%s/.AppleDouble/%s", path, file);
    if (!(ainfo = fopen (buf, "r"))) {
      /* Can't open? Just set some defaults then.
	 Really, it should pop a panel and ask, but since
	 this will be called from the CPU thread, we can't */
      info.fileType = 0x06; /* BIN */
      info.auxType = 0x00;
    }
    else { /* Make this read apple double later */
      fscanf (ainfo, "%X\n%X\n", &info.fileType, &info.auxType);
      fclose (ainfo);
    }
  }

  return &info;
}

int setAppleDoubleInfo (char *filename, fileInfo *finfo)
{
  char path[MAXPATHLEN];
  char file[MAXPATHLEN];
  char buf[MAXPATHLEN];
  struct stat sbuf;
  FILE *ainfo;
  char *p;

  int mkdir ();
  

  strcpy (path, filename);
  if (p = rindex (path, '/')) {
    strcpy (file, p+1);
    *p = 0;
  }
  else {
    path[0] = 0;
    strcpy (file, filename);
  }

  if (finfo->fileType != 0x0f) {
    sprintf (buf, "%s/.AppleDouble", path);
    if (stat (buf, &sbuf)) {
      if (mkdir (buf, 0777))
	return 0;
    }
    
    sprintf (buf, "%s/.AppleDouble/%s", path, file);
    if (!(ainfo = fopen (buf, "w")))
      return 0;
    
    fprintf (ainfo, "%X\n%X\n", finfo->fileType, finfo->auxType);
    fclose (ainfo);
    
    sprintf (buf, "%s/%s", path, file);
  }

  /* About all we can do is try to change permissions.
     I don't see any way to change the date. Oh well. */
  /* Well, actually, I guess utime lets you change
     modify and access times. */
  /* Also, I've noticed that on the NeXT, the create time
     is almost always wrong. */
  /* I'll set it up later */
  
  return 1;
}

int removeAppleDoubleInfo (char *filename)
{
  char path[MAXPATHLEN];
  char file[MAXPATHLEN];
  char buf[MAXPATHLEN];
  struct stat sbuf;
  char *p;

  int unlink ();
  
  
  strcpy (path, filename);
  if (p = rindex (path, '/')) {
    strcpy (file, p+1);
    *p = 0;
  }
  else {
    path[0] = 0;
    strcpy (file, filename);
  }

  sprintf (buf, "%s/.AppleDouble", path);
  if (stat (buf, &sbuf)) {
    return 0;
  }

  sprintf (buf, "%s/.AppleDouble/%s", path, file);
  if (!stat (buf, &sbuf))
    unlink (buf);
    
  return 1;
}

char *makeCString (ADDR fname)
{
  static char pname[129];
  int i, j;

  
  j = READLOC (fname);
  for (i = 0; i < j; i++)
    pname[i] = LOWER (READLOC (fname + i + 1) & 0x7f);
  pname[i] = 0;

  return pname;
}

void writePString (ADDR where, char *fname)
{
  int i;

  
  WRITELOC (where, strlen (fname));

  for (i = 0; fname[i]; i++)
    WRITELOC (where + 1 + i, UPPER (fname[i]));
}

int validPName (char *path)
{
  int i, j;
  

//  fprintf (stderr, "Path: %s\n", path);
  
  for (i = 0, j = 1; path[i]; i++) {
    if (path[i] != '/') {
      if (j && ((UPPER(path[i]) < 'A' || UPPER(path[i]) > 'Z')))
	return 0;
      j = UPPER(path[i]);
      if (j < '.' || (j > '.' && j < '0') || (j > '9' && j < 'A') || j > 'Z')
	return 0;
      j = 0;      
    }
    else
      j = 1;
  }

//  fprintf (stderr, "valid\n");
  
  return 1;
}
    
int mapPName (char *fname)
     /* This will read through the directory to find a match
	for fname. If it finds one, it will convert fname to it */
{
  static char path[MAXPATHLEN];
  static char file[MAXPATHLEN];
  char *p;
  DIR *dirp;
  struct direct *ent;
  int len;


  strcpy (path, fname);
  if (p = rindex (path, '/')) {
    strcpy (file, p+1);
    *p = 0;
  }
  else {
    strcpy (path, ".");
    strcpy (file, fname);
  }

  if (!(dirp = opendir (path)))
    return 0;

  len = strlen(file);

  for (ent = readdir(dirp); ent; ent = readdir(dirp))
    if (ent->d_namlen == len && validPName (ent->d_name) && !strcasecmp (ent->d_name, file)) {
      closedir(dirp);
      if (p = rindex (fname, '/'))
	p++;
      else
	p = fname;
      strcpy (p, ent->d_name);
      return 1;
    }
  
  closedir(dirp);
  
  return 0;
}
  
char *makeUnixPath (ADDR fname)
     /* This returns NULL if the path won't be in our UNIX dir.
	fname is a pointer to the actual Prodos filename in
	SysMem. */
{
  static char fullname[MAXPATHLEN];
  char pname[129];
  char *p;
  int i, j;


  j = READLOC (fname);
  for (i = 0; i < j; i++) {
    pname[i] = READLOC (fname + i + 1) & 0x7f;
    pname[i] = pname[i] >= 'A' && pname[i] <= 'Z' ? pname[i] + ' ' : pname[i];
  }
  pname[i] = 0;

  if (pname[0] != '/') {
    strcpy (fullname, pname);
    sprintf (pname, "%s%s", unixPrefix, fullname);
  }
  
  if (pname[0] == '/') {
    if (!validPName (pname))
      return NULL;
    
    strcpy (fullname, &pname[1]);
    if (p = index (fullname, '/'))
      *p = 0;
    if (strcasecmp (fullname, unixVolumeName))
      return NULL;
    if (p = index (&pname[1], '/')) {
      p++;
      sprintf (fullname, "%s/%s/%s", unixVolumePath, unixVolumeName, p);
    }
    else
      sprintf (fullname, "%s/%s", unixVolumePath, unixVolumeName);
  }
  else {
    return NULL;
  }

  mapPName (fullname);
    
  return fullname;
}

char *loadDir (char *path, int *length)
     /* This will load an entire directory into ram, converting into
	the ProDOS directory structure. */
{
  char *dirBlock;
  DIR *dirp;
  struct direct *ent;
  int i, j, k, l;
  fileInfo *finfo;
  char buf[MAXPATHLEN];
  char *p;

  void *malloc ();


  if (!(dirp = opendir (path)))
    return NULL;

  /* Find out how many entries there are */
  k = 43;
  i = 1;
  while (ent = readdir (dirp)) {
    if (validPName (ent->d_name)) {
      i++;
      k += 0x27;
      if ((512 - k % 512) < 0x27)
	k += (512 - k % 512) + 4;
    }
  }
  rewinddir (dirp);

  dirBlock = malloc (k);
  *length = k;

  strcpy (buf, path);
  if (buf[strlen (buf) - 1] == '/')
    buf[strlen (buf) - 1] = 0;
  
  if (p = rindex (buf, '/'))
    p++;
  else
    p = path;

  k = 4;
  finfo = getAppleDoubleInfo (path);
  dirBlock[k + 0x00] = (finfo->storageType << 4) | (strlen (p) <= 15 ? strlen (p) : 15);
  dirBlock[k + 0x00] += 0x10;
  for (l = 0; l < (dirBlock[k + 0x00] & 0xf); l++)
    dirBlock[k + 0x01 + l] = UPPER(p[l]);
  dirBlock[k + 0x10] = 0x75;
  dirBlock[k + 0x18] = finfo->dateCreate & 0xff;
  dirBlock[k + 0x19] = (finfo->dateCreate >> 8) & 0xff;
  dirBlock[k + 0x1a] = finfo->timeCreate & 0xff;
  dirBlock[k + 0x1b] = (finfo->timeCreate >> 8) & 0xff;
  dirBlock[k + 0x1c] = 0;
  dirBlock[k + 0x1d] = 0;
  dirBlock[k + 0x1e] = finfo->accessBits;
  dirBlock[k + 0x1f] = 0x27;
  dirBlock[k + 0x20] = 0x0d;
  dirBlock[k + 0x21] = (i - 1) & 0xff;
  dirBlock[k + 0x22] = ((i - 1) >> 8) & 0xff;
  dirBlock[k + 0x23] = dirBlock[k + 0x24] = dirBlock[k + 0x25] = dirBlock[k + 0x26] = 0;
  k += 0x27;
  
  for (j = 1; j < i;) {
    ent = readdir (dirp);
    if (validPName (ent->d_name)) {
      if ((512 - k % 512) < 0x27)
	k += (512 - k % 512) + 4;
      sprintf (buf, "%s/%s", path, ent->d_name);
      finfo = getAppleDoubleInfo (buf);
      dirBlock[k + 0x00] = (finfo->storageType << 4) | (ent->d_namlen <= 15 ? ent->d_namlen : 15);
      for (l = 0; l < (dirBlock[k + 0x00] & 0xf); l++)
	dirBlock[k + 0x01 + l] = UPPER(ent->d_name[l]);
      dirBlock[k + 0x10] = finfo->fileType;
      dirBlock[k + 0x11] = dirBlock[k + 0x12] = 0;
      dirBlock[k + 0x13] = finfo->blocksUsed & 0xff;
      dirBlock[k + 0x14] = (finfo->blocksUsed >> 8) & 0xff;
      dirBlock[k + 0x15] = finfo->fileLength & 0xff;
      dirBlock[k + 0x16] = (finfo->fileLength >> 8) & 0xff;
      dirBlock[k + 0x17] = (finfo->fileLength >> 16) & 0xff;
      dirBlock[k + 0x18] = finfo->dateCreate & 0xff;
      dirBlock[k + 0x19] = (finfo->dateCreate >> 8) & 0xff;
      dirBlock[k + 0x1a] = finfo->timeCreate & 0xff;
      dirBlock[k + 0x1b] = (finfo->timeCreate >> 8) & 0xff;
      dirBlock[k + 0x1c] = 0;
      dirBlock[k + 0x1d] = 0;
      dirBlock[k + 0x1e] = finfo->accessBits;
      dirBlock[k + 0x1f] = finfo->auxType & 0xff;
      dirBlock[k + 0x20] = (finfo->auxType >> 8) & 0xff;
      dirBlock[k + 0x21] = finfo->dateLastMod & 0xff;
      dirBlock[k + 0x22] = (finfo->dateLastMod >> 8) & 0xff;
      dirBlock[k + 0x23] = finfo->timeLastMod & 0xff;
      dirBlock[k + 0x24] = (finfo->timeLastMod >> 8) & 0xff;
      dirBlock[k + 0x25] = dirBlock[k + 0x26] = 0x00;
      j++;
      k += 0x27;
    }
  }
  closedir (dirp);

  return dirBlock;
}

int mapErrnoToPdos (int err)
     /* This is very incomplete. Some of these may be
      better suited to different Prodos errors as well. */
{
  switch (err) {
  case EPERM:
    return PDOSERR_ACCESSERROR;
  case ENOENT:
    return PDOSERR_FILENOTFOUND;
  case EIO:
    return PDOSERR_IOERROR;
  case ENXIO:
    return PDOSERR_IOERROR;
  case EBADF:
    return PDOSERR_IOERROR;
  case EACCES:
    return PDOSERR_ACCESSERROR;
  case EEXIST:
    return PDOSERR_DUPFILENAME;
  case ENOTDIR:
    return PDOSERR_NOTADIR;
  case EISDIR:
    return PDOSERR_ACCESSERROR;
  case EMFILE:
    return PDOSERR_TOOMANYOPEN;
  case ETXTBSY:
    return PDOSERR_IOERROR;
  case EFBIG:
    return PDOSERR_IOERROR;
  case ENOSPC:
    return PDOSERR_DISKFULL;
  case EROFS:
    return PDOSERR_ACCESSERROR;
  default:
    return PDOSERR_IOERROR;
  }

  return 0;
}

int unixOpen (char *path)
     /* This will handle opens on directories as well as files */
{
  fileInfo *finfo;
  int i;
  

  pdoserrno = 0;
  
  for (i = 0; unixFile[i]; i++);
  if (i > 7) {
    pdoserrno = PDOSERR_TOOMANYOPEN;
    return 0;
  }    

  if (finfo = getAppleDoubleInfo (path)) {
    if (finfo->fileType == 0x0f) { /* Opening a directory? */
      unixFile[i] = (FILE *) loadDir (path, &unixFileLength[i]); /* Address of dir buf */
      unixFileMode[i] = UNIX_DIRECTORY | READLOC (0xbf94);
      unixFilePos[i] = 0;
      unixFileNewlineMask[i] = 0;
    }
    else { /* Just a file */
      if (!(unixFile[i] = fopen (path, "r+"))) { /* Always opens r/w. Should fix it */
	pdoserrno = PDOSERR_IOERROR; /* does global errno get set? */
	return 0;
      }
      unixFileMode[i] = READLOC (0xbf94);
      unixFileLength[i] = finfo->fileLength;
      unixFileNewlineMask[i] = 0;
    }
  }
  else {
    pdoserrno = mapErrnoToPdos (errno);
    return 0;
  }

  return i + 1;
}

int unixSetNewline (int fnum, int nmask, int nchar)
{
  pdoserrno = 0;

  if (fnum < 1 || fnum > 8 || !unixFile[fnum - 1]) {
    pdoserrno = PDOSERR_INVALIDREFNUM;
    return 0;
  }
  
  unixFileNewlineMask[fnum - 1] = nmask;
  unixFileNewlineChar[fnum - 1] = nchar;

  return fnum;
}
  
static BYTE readbuf[0xffff];

int unixRead (int fnum, int length, int bufaddr)
     /* This handles reads from unix files into SysMem,
	including directories and newline mapping */
{
  int i, j;
  char *p;
  

  pdoserrno = 0;
  
  if (fnum < 1 || fnum > 8 || !unixFile[fnum - 1]) {
    pdoserrno = PDOSERR_INVALIDREFNUM;
    return 0;
  }
  
  if (unixFileMode[fnum - 1] & UNIX_DIRECTORY) {
    i = length;
    if (i + unixFilePos[fnum - 1] > unixFileLength[fnum - 1])
      i = unixFileLength[fnum - 1] - unixFilePos[fnum - 1];
    p = (char *) unixFile[fnum - 1];
    bcopy ((char *) &p[unixFilePos[fnum - 1]], readbuf, i);

    for (j = 0; j < i; j++) {
      WRITELOC (bufaddr + j, readbuf[j]);
      if (unixFileNewlineMask[fnum - 1] && (i & unixFileNewlineMask[fnum - 1]) == unixFileNewlineChar[fnum - 1]) {
	j++;
	break;
      }
    }
    
    unixFilePos[fnum - 1] += j;
  }
  else {
    for (j = 0; j < length; j++) {
      i = fgetc (unixFile[fnum - 1]);
      if (i == EOF)
	break;
      WRITELOC (bufaddr + j, i);
      if (unixFileNewlineMask[fnum - 1] && (i & unixFileNewlineMask[fnum - 1]) == unixFileNewlineChar[fnum - 1]) {
	j++;
	break;
      }
    }
  }
  
  if (!j) {
    pdoserrno = PDOSERR_ENDOFFILE;
    return 0;
  }
  
  return j;
}

int unixWrite (int fnum, int length, int bufaddr)
     /* This handles reads from unix files into SysMem,
	including directories and newline mapping */
{
  int i, j;
  

  pdoserrno = 0;
  
  if (fnum < 1 || fnum > 8 || !unixFile[fnum - 1]) {
    pdoserrno = PDOSERR_INVALIDREFNUM;
    return 0;
  }
  
  if (unixFileMode[fnum - 1] & UNIX_DIRECTORY) {
    pdoserrno = PDOSERR_ACCESSERROR;
    return 0;
  }
  else {
    for (i = 0; i < length; i++)
      readbuf[i] = READLOC (bufaddr + i);
  }

  j = fwrite (readbuf, 1, length, unixFile[fnum - 1]);
  
  return j;
}

int unixClose (int fnum)
{
  int i, j;

  void free ();

  
  pdoserrno = 0;

  if (!fnum) {
    j = READLOC (0xbf94);
    for (i = 0; i < 8; i++)
      if ((unixFileMode[i] & 0xff) >= j && unixFile[i])
	unixClose (i+1);
    return -1;
  }
  else {
    if (fnum < 1 || fnum > 8 || !unixFile[fnum - 1]) {
      pdoserrno = PDOSERR_INVALIDREFNUM;
      return 0;
    }
    
    if (unixFileMode[fnum - 1] & UNIX_DIRECTORY) {
      free (unixFile[fnum - 1]);
    }
    else {
      fclose (unixFile[fnum - 1]);
    }
    
    unixFileMode[fnum - 1] = 0;
    unixFile[fnum - 1] = NULL;
  }

  return fnum;
}

int unixSetMark (int fnum, int where)
{
  int i;

  
  pdoserrno = 0;
  
  if (unixFileMode[fnum - 1] & UNIX_DIRECTORY)
    unixFilePos[fnum - 1] = where;
  else {
    if (fseek (unixFile[fnum - 1], where, SEEK_SET)) {
      pdoserrno = mapErrnoToPdos (errno);
      return 0;
    }
    i = ftell (unixFile[fnum - 1]);
    if (i != where) {
      pdoserrno = PDOSERR_IOERROR;
      return 0;
    }
  }
  
  return fnum;
}

int unixGetMark (int fnum)
{
  pdoserrno = 0;
  
  if (unixFileMode[fnum - 1] & UNIX_DIRECTORY)
    return unixFilePos[fnum - 1];
  else
    return ftell (unixFile[fnum - 1]);

  return fnum;
}

void printparms (ADDR addr, char *pat)
{
  int i, j, k, l;


  for (i = 0; pat[i]; i++) {
    switch (pat[i]) {
    case 'b':
      fprintf (stderr, "%02X\n", READLOC (addr));
      addr++;
      break;
    case 'w':
      fprintf (stderr, "%04X\n", READLOC (addr) | (READLOC(addr+1) << 8));
      addr += 2;
      break;
    case '3':
      fprintf (stderr, "%06X\n", READLOC (addr) | (READLOC(addr+1) << 8) | (READLOC(addr+2)<<16));
      addr += 3;
      break;
    case 'p':
      fprintf (stderr, "%04X->", j = READLOC (addr) | (READLOC(addr+1) << 8));
      fprintf (stderr, "%02X ", k = READLOC(j));
      for (l = 0; l < k; l++)
	fprintf (stderr, "%c", READLOC(j + 1 + l) & 0x7f);
      fprintf (stderr, "\n");
    }
  }

  return;
}

void unixShare ()
{
  int retAddr = READWORD (0x100 + Sp + 1);
  int cmd = READLOC (retAddr + 1), parms = READWORD (retAddr + 2);
  char *up;
  char up2[MAXPATHLEN];
  fileInfo finfo;
  fileInfo *finfo2;
  FILE *file;
  int i, j, l;

  int mkdir ();
  int unlink ();

  extern int BreakPoint;
  

  lastCmd = cmd;
  /* Move return address past parameters */
  retAddr += 3;
  WRITELOC (0x100 + Sp + 1, retAddr & 0xff);
  WRITELOC (0x100 + Sp + 2, (retAddr >> 8) & 0xff);

  /* Will do JSR to Prodos inside us */
  SysMem[UNIX_COMMAND] = cmd;
  SysMem[UNIX_PARM] = parms & 0xff;
  SysMem[UNIX_PARM + 1] = (parms >> 8) & 0xff;

  if (BreakPoint)
    fprintf (stderr, "Command: %02X\nParms: %04X\n", cmd, parms);
  
  N = 0; /* If negative is set, not us, pass it on to Prodos */
  
  switch (cmd) {
  case 0x80:
    if (BreakPoint)
      printparms (parms, "bbww");
    if (READLOC(parms + 1) == UNIX_SLOTDRIVE) {
      C = 1;
      LDA(PDOSERR_NODEVICECNCTD);
      return;
    }
    else
      N = 1;
    return;
    break;
  case 0x81:
    if (BreakPoint)
      printparms (parms, "bbww");
    if (READLOC(parms + 1) == UNIX_SLOTDRIVE) {
      C = 1;
      LDA(PDOSERR_NODEVICECNCTD);
      return;
    }
    else
      N = 1;
    return;
    break;
  case 0xc0: /* Create */
    if (BreakPoint)
      printparms (parms, "bpbbwbww");
    if (up = makeUnixPath (READWORD(parms + 1))) {
      if (getAppleDoubleInfo (up)) { /* Can't exist */
	C = 1;
	LDA(PDOSERR_DUPFILENAME);
	return;
      }
      else {
	finfo.accessBits = READLOC (parms + 3);
	finfo.fileType = READLOC (parms + 4);
	finfo.auxType = READWORD (parms + 5);
	finfo.storageType = READLOC (parms + 7);
	finfo.dateCreate = READWORD (parms + 8);
	finfo.timeCreate = READWORD (parms + 10);
	if (finfo.fileType == 0x0f) { /* Directory */
	  if (mkdir (up, 0777)) {
	    C = 1;
	    LDA(mapErrnoToPdos (errno));
	    return;
	  }
	  setAppleDoubleInfo (up, &finfo);
	  C = 0;
	  LDA(0);
	}
	else {
	  if (!(file = fopen (up, "w"))) {
	    C = 1;
	    LDA(PDOSERR_IOERROR);
	    return;
	  }
	  fclose (file);
	  setAppleDoubleInfo (up, &finfo);
	  C = 0;
	  LDA(0);
	}
      }
    }
    else
      N = 1;
    return;
    break;
  case 0xc1: /* Destroy */
    if (BreakPoint)
      printparms (parms, "bp");
    if (up = makeUnixPath (READWORD(parms + 1))) {
      if (unlink (up)) {
	C = 1;
	LDA(mapErrnoToPdos (errno));
	return;
      }
      removeAppleDoubleInfo (up);
      C = 0;
      LDA(0);
    }
    else
      N = 1;
    return;
    break;
  case 0xc2: /* Rename */
    if (BreakPoint)
      printparms (parms, "bpp");
    if (up = makeUnixPath (READWORD(parms + 1))) {
      strcpy (up2, up);
      if (up = makeUnixPath (READWORD(parms + 3))) {
	if (rename (up2, up)) {
	  C = 1;
	  LDA(mapErrnoToPdos (errno));
	  return;
	}
      }
      else {
	C = 1;
	LDA(PDOSERR_INVALIDPATHSYN);
	return;
      }
      C = 0;
      LDA(0);
    }
    else
      N = 1;
    return;
    break;
  case 0xc3: /* Set File Info */
    if (BreakPoint)
      printparms (parms, "bpbbwbwww");
    if (up = makeUnixPath (READWORD(parms + 1))) {
      finfo.accessBits = READLOC (parms + 3);
      finfo.fileType = READLOC (parms + 4);
      finfo.auxType = READWORD (parms + 5);
      finfo.dateLastMod = READWORD (parms + 10);
      finfo.timeLastMod = READWORD (parms + 12);
      setAppleDoubleInfo (up, &finfo);
      C = 0;
      LDA(0);
    }
    else
      N = 1;
    return;
    break;
  case 0xc4: /* Get File Info */
    if (BreakPoint)
      printparms (parms, "bp");
    if (up = makeUnixPath (READWORD(parms + 1))) {
      if (finfo2 = getAppleDoubleInfo (up)) {
	WRITELOC (parms + 3, finfo2->accessBits);
	WRITELOC (parms + 4, finfo2->fileType);
	WRITELOC (parms + 5, finfo2->auxType & 0xff);
	WRITELOC (parms + 6, (finfo2->auxType >> 8) & 0xff);
	WRITELOC (parms + 7, finfo2->storageType);
	WRITELOC (parms + 8, finfo2->blocksUsed & 0xff);
	WRITELOC (parms + 9, (finfo2->blocksUsed >> 8) & 0xff);
	WRITELOC (parms + 10, finfo2->dateLastMod & 0xff);
	WRITELOC (parms + 11, (finfo2->dateLastMod >> 8) & 0xff);
	WRITELOC (parms + 12, finfo2->timeLastMod & 0xff);
	WRITELOC (parms + 13, (finfo2->timeLastMod >> 8) & 0xff);
	WRITELOC (parms + 14, finfo2->dateCreate & 0xff);
	WRITELOC (parms + 15, (finfo2->dateCreate >> 8) & 0xff);
	WRITELOC (parms + 16, finfo2->timeCreate & 0xff);
	WRITELOC (parms + 17, (finfo2->timeCreate >> 8) & 0xff);
      }
      else {
	C = 1;
	LDA(mapErrnoToPdos (errno));
	return;
      }
      C = 0;
      LDA(0);
      if (BreakPoint)
	printparms (parms + 3, "bbwbwwwww");
    }
    else
      N = 1;
    return;
    break;
  case 0xc5: /* Online */
    /* What does AppleShare do about this?? */
    if (BreakPoint)
      printparms (parms, "bbw");
/*    if (!READLOC(parms + 1)) {
      oldAddress = READWORD(parms + 2);
      writePString (oldAddress, unixVolumeName);
      WRITELOC (oldAddress, READLOC (oldAddress) | UNIX_SLOTDRIVE);
      WRITELOC(parms + 2, (oldAddress + 16) & 0xff);
      WRITELOC(parms + 3, ((oldAddress + 16) >> 8) & 0xff);
      N = 1;
    }
    else */ if (READLOC(parms + 1) == UNIX_SLOTDRIVE) {
      j = READWORD(parms + 2);
      writePString (j, unixVolumeName);
      WRITELOC (j, READLOC (j) | UNIX_SLOTDRIVE);

      C = 0;
      LDA(0);
    }
    else 
      N = 1;
    return;
    break;
  case 0xc6: /* Set Prefix */
    if (BreakPoint)
      printparms (parms, "bp");
    if (up = makeUnixPath (READWORD(parms + 1))) {
      if (!(finfo2 = getAppleDoubleInfo (up))) {
	C = 1;
	LDA(PDOSERR_NOTADIR);
	return;
      }
      if (finfo2->fileType != 0x0f) {
	C = 1;
	LDA(PDOSERR_INCOMPFILETYPE);
	return;
      }
      up = makeCString (READWORD(parms + 1));
      if (up[0] == '/')
	strcpy (unixPrefix, up);
      else
	strcat (unixPrefix, up);
      if (unixPrefix[strlen (unixPrefix) - 1] != '/')
	strcat (unixPrefix, "/");
      WRITELOC (0xbf9a, 0xff);
      C = 0;
      LDA(0);
    }
    else {
      /* If prodos changes it, then unset us. */
      N = 1;
    }      
    return;
    break;
  case 0xc7: /* Get Prefix */
    if (BreakPoint)
      printparms (parms, "b");
    if (unixPrefix[0]) {
      writePString (READWORD(parms + 1), unixPrefix);
      C = 0;
      LDA(0);
      if (BreakPoint)
	printparms (parms + 1, "p");
    }
    else
      N = 1;
    return;
    break;
  case 0xc8: /* Open */
    if (BreakPoint)
      printparms (parms, "bpw");
    if (up = makeUnixPath (READWORD(parms + 1))) {
      if (!(i = unixOpen (up))) {
	C = 1;
	LDA(pdoserrno);
	return;
      }
      WRITELOC (parms + 5, i | 0x80);
      C = 0;
      LDA(0);
      if (BreakPoint)
	printparms (parms + 5, "b");
    }
    else
      N = 1;
    return;
    break;
  case 0xc9: /* Newline */
    if (BreakPoint)
      printparms (parms, "bbbb");
    if ((j = READLOC (parms + 1)) & 0x80) {
      j &= 0x7f;
      unixSetNewline (j, READLOC(parms + 2), READLOC(parms + 3));
      C = 0;
      LDA(0);
    }
    else
      N = 1;
    return;
    break;
  case 0xca: /* Read */
    if (BreakPoint)
      printparms (parms, "bbww");
    if ((j = READLOC (parms + 1)) & 0x80) {
      j &= 0x7f;
      l = unixRead (j, READWORD (parms + 4), READWORD (parms + 2));
      WRITELOC (parms + 6, low(l));
      WRITELOC (parms + 7, high(l));
      if (!l) {
	C = 1;
	LDA(pdoserrno);
	return;
      }
      C = 0;
      LDA(0);
      if (BreakPoint)
	printparms (parms + 6, "w");
    }
    else
      N = 1;
    return;
    break;
  case 0xcb: /* Write */
    if (BreakPoint)
      printparms (parms, "bbww");
    if ((j = READLOC (parms + 1)) & 0x80) {
      j &= 0x7f;
      l = unixWrite (j, READWORD (parms + 4), READWORD (parms + 2));
      WRITELOC (parms + 6, low(l));
      WRITELOC (parms + 7, high(l));
      if (!l) {
	C = 1;
	LDA(pdoserrno);
	return;
      }
      C = 0;
      LDA(0);
      if (BreakPoint)
	printparms (parms + 6, "w");
    }
    else
      N = 1;
    return;
    break;
  case 0xcc: /* Close */
    if (BreakPoint)
      printparms (parms, "bb");
    j = READLOC (parms + 1);
    if ((!j) || (j & 0x80)) {
      j &= 0x7f;
      i = unixClose (j);
      if (!i) {
	C = 1;
	LDA(pdoserrno);
	return;
      }
      if (i < 0) {
	N = 1;
	return;
      }
      C = 0;
      LDA(0);
    }
    else 
      N = 1;
    return;
    break;
  case 0xcd: /* Flush */
    if (BreakPoint)
      printparms (parms, "bb");
    if ((j = READLOC (parms + 1)) & 0x80) { /* could but why bother */
      C = 0;
      LDA(0);
    }
    else
      N = 1;
    return;
    break;
  case 0xce: /* Set Mark */
    if (BreakPoint)
      printparms (parms, "bb3");
    if ((j = READLOC (parms + 1)) & 0x80) {
      j &= 0x7f;
      if (!unixSetMark (j, READWORD (parms + 2) | (READLOC (parms + 4) << 16))) {
	C = 1;
	LDA(pdoserrno);
	return;
      }
      C = 0;
      LDA(0);
    }
    else
      N = 1;
    return;
    break;
  case 0xcf: /* Get Mark */
    if (BreakPoint)
      printparms (parms, "bb");
    if ((j = READLOC (parms + 1)) & 0x80) {
      j &= 0x7f;
      i = unixGetMark (j);
      WRITELOC (parms + 2, low (i));
      WRITELOC (parms + 3, high (i));
      WRITELOC (parms + 4, i >> 16);
      C = 0;
      LDA(0);
      if (BreakPoint)
	printparms (parms + 2, "3");
    }
    else
      N = 1;
    return;
    break;
  case 0xd0: /* Set EOF */
    if (BreakPoint)
      printparms (parms, "bb3");
    if ((j = READLOC (parms + 1)) & 0x80) { /* How do I do this? */
      C = 0;
      LDA(0);
    }
    else
      N = 1;
    return;
    break;
  case 0xd1: /* Get EOF */
    if (BreakPoint)
      printparms (parms, "bb");
    if ((j = READLOC (parms + 1)) & 0x80) {
      j &= 0x7f;
      WRITELOC (parms + 2, unixFileLength[j - 1]);
      WRITELOC (parms + 3, unixFileLength[j - 1] >> 8);
      WRITELOC (parms + 4, unixFileLength[j - 1] >> 16);
      
      C = 0;
      LDA(0);
      if (BreakPoint)
	printparms (parms + 2, "3");
    }
    else
      N = 1;
    return;
    break;
  case 0xd2: /* Set Buf */
    if (BreakPoint)
      printparms (parms, "bbw");
    if ((j = READLOC (parms + 1)) & 0x80) {
      C = 0;
      LDA(0);
    }
    else
      N = 1;
    return;
    break;
  case 0xd3: /* Get Buf */
    if (BreakPoint)
      printparms (parms, "bbw");
    if ((j = READLOC (parms + 1)) & 0x80) {
      C = 0;
      LDA(0);
    }
    else
      N = 1;
    return;
    break;
  default: /* We don't want it. Let Prodos handle it */
    N = 1;
    return;
    break;
  }
    
  return;
}

void unixReturnVal ()
{
  extern int BreakPoint;

  
  if (BreakPoint)
    fprintf (stderr, "Prodos returned: %02X\n", A);
  
  if (lastCmd == 0xc6 && !C && !A)
    unixPrefix[0] = 0;
  if (lastCmd == 0xc5 && oldAddress) {
    WRITELOC (READWORD (UNIX_PARM), oldAddress & 0xff);
    WRITELOC (READWORD (UNIX_PARM) + 1, (oldAddress >> 8) & 0xff);
    oldAddress = 0;
  }
  lastCmd = 0;
  
  return;
}
